
import unittest
import numpy as np
from uhepp import UHeppHist, Yield, Graph

class NumpyTestCase(unittest.TestCase):
    """Test that assigned values are converted to plain python"""

    def test_edges(self):
        """Check assigning a numpy array to edges converts it to python"""
        edges = np.linspace(0, 100, 11)
        hist = UHeppHist("m", edges)

        self.assertIsInstance(hist.bin_edges, list) 
        self.assertIsInstance(hist.bin_edges[0], (int, float)) 
        self.assertNotIsInstance(hist.bin_edges[0], (np.float64, np.int64))

    def test_base(self):
        """Check assigning a numpy array to base yields converts it to python"""
        base = np.arange(10)**2
        yield_ = Yield(base)

        self.assertIsInstance(yield_.base, list)
        self.assertIsInstance(yield_.base[0], (int, float)) 
        self.assertNotIsInstance(yield_.base[0], (np.float64, np.int64))

    def test_stat(self):
        """Check assigning a numpy array to stat yields converts it to python"""
        base = np.arange(10)**2
        stat = np.ones(10) * 8
        yield_ = Yield(base, stat)

        self.assertIsInstance(yield_.stat, list)
        self.assertIsInstance(yield_.stat[0], (int, float)) 
        self.assertNotIsInstance(yield_.stat[0], (np.float64, np.int64))

    def test_syst(self):
        """Check assigning a numpy array to syst yields converts it to python"""
        base = np.arange(10)**2
        syst = np.ones(10) * 8
        yield_ = Yield(base, syst=syst)

        self.assertIsInstance(yield_.syst, list)
        self.assertIsInstance(yield_.syst[0], (int, float)) 
        self.assertNotIsInstance(yield_.syst[0], (np.float64, np.int64))

    def test_var_up(self):
        """Check assigning a numpy array to var yields converts it to python"""
        base = np.arange(10)**2
        var = base + np.ones(10) * 8
        yield_ = Yield(base)
        yield_.add_var("DETECTOR", var_up=var, var_down=var)

        self.assertIsInstance(yield_.var_up("DETECTOR"), list)
        self.assertIsInstance(yield_.var_up("DETECTOR")[0], (int, float)) 
        self.assertNotIsInstance(yield_.var_up("DETECTOR")[0],
                                 (np.float64, np.int64))

    def test_var_down(self):
        """Check assigning a numpy array to var yields converts it to python"""
        base = np.arange(10)**2
        var = base + np.ones(10) * 8
        yield_ = Yield(base)
        yield_.add_var("DETECTOR", var_up=var, var_down=var)

        self.assertIsInstance(yield_.var_down("DETECTOR"), list)
        self.assertIsInstance(yield_.var_down("DETECTOR")[0], (int, float)) 
        self.assertNotIsInstance(yield_.var_down("DETECTOR")[0],
                                 (np.float64, np.int64))


    def test_graph_x(self):
        """Check assigning a numpy array to graph converts it to python"""
        series = np.arange(5) / 5
        graph = Graph(series, series)

        self.assertIsInstance(graph.x_values, list)
        self.assertIsInstance(graph.x_values[0], (int, float)) 
        self.assertNotIsInstance(graph.x_values[0], (np.float64, np.int64))

    def test_graph_y(self):
        """Check assigning a numpy array to graph converts it to python"""
        series = np.arange(5) / 5
        graph = Graph(series, series)

        self.assertIsInstance(graph.y_values, list)
        self.assertIsInstance(graph.y_values[0], (int, float)) 
        self.assertNotIsInstance(graph.y_values[0], (np.float64, np.int64))

    def test_graph_xerr(self):
        """Check assigning a numpy array to graph converts it to python"""
        series = np.arange(5) / 5
        graph = Graph(series, series)
        graph.x_errors = series
        graph.y_errors = series

        self.assertIsInstance(graph.x_errors, list)
        self.assertIsInstance(graph.x_errors[0], (int, float)) 
        self.assertNotIsInstance(graph.x_errors[0], (np.float64, np.int64))

    def test_graph_yerr(self):
        """Check assigning a numpy array to graph converts it to python"""
        series = np.arange(5) / 5
        graph = Graph(series, series)
        graph.x_errors = series
        graph.y_errors = series

        self.assertIsInstance(graph.y_errors, list)
        self.assertIsInstance(graph.y_errors[0], (int, float)) 
        self.assertNotIsInstance(graph.y_errors[0], (np.float64, np.int64))




import unittest
import numpy as np
from uhepp import StackItem

class StackItemTestCase(unittest.TestCase):
    """Test the implementation of StackItem"""

    def test_init_store(self):
        """Check that arguments to init are stored as members"""
        stackitem = StackItem(["ggh", "vbf"], label="Signal", color='k')

        self.assertEqual(stackitem.yield_names, ['ggh', 'vbf'])
        self.assertEqual(stackitem.label, 'Signal')
        self.assertEqual(stackitem.color, '#000000')

    def test_init_copy(self):
        """Check that arguments to init are copied"""
        yields = ['ggh', 'vbf']
        stackitem = StackItem(yields, label="Signal", color="k")

        yields.append('43')

        self.assertEqual(stackitem.yield_names, ['ggh', 'vbf'])
        self.assertEqual(stackitem.label, 'Signal')
        self.assertEqual(stackitem.color, '#000000')

    def test_member_change(self):
        """Check that member variables can be changed"""
        stackitem = StackItem(["ggh", "vbf"], label="Signal", color='#000000')

        stackitem.yield_names.append('tth')
        stackitem.color = 'r'

        self.assertEqual(stackitem.yield_names, ['ggh', 'vbf', 'tth'])
        self.assertEqual(stackitem.label, 'Signal')
        self.assertEqual(stackitem.color, "#ff0000")

    def test_illegal_style(self):
        """Check that an illegal style key raises an exception"""
        self.assertRaises(ValueError, StackItem, ['ggh'],
                          label="Signal", whaterever=32)

    def test_color_rgb(self):
        """Check that a 3-tuple is converted to a hex string"""
        self.assertEqual(StackItem._parse_color((0.1, 0.9, 0)), "#1ae600")

    def test_color_rgba(self):
        """Check that a 4-tuple is converted to a hex string"""
        self.assertEqual(StackItem._parse_color((0.1, 0.9, 0, 0.1)),
                         "#1ae6001a")

    def test_line_style_legal(self):
        """Check that legal line styles are accepted"""
        StackItem(["ggh", "vbf"], "Signal", linestyle='--')
        StackItem(["ggh", "vbf"], "Signal", linestyle='-.')
        StackItem(["ggh", "vbf"], "Signal", linestyle=':')
        StackItem(["ggh", "vbf"], "Signal", linestyle='-')

    def test_line_style_illegal(self):
        """Check that invalid line styles raise an exception"""
        self.assertRaises(ValueError, StackItem, ["ggh", "vbf"], "Signal",
                          linestyle='+')

    def test_line_width(self):
        """Check that line widths are converted to floats"""
        stackitem = StackItem(["ggh", "vbf"], "Signal", linewidth='3')

        self.assertEqual(stackitem.linewidth, 3)
        self.assertIsInstance(stackitem.linewidth, float)

    def test_default_style(self):
        """Check that the default styles are None"""
        stackitem = StackItem(["ggh", "vbf"], "Signal")

        self.assertIsNone(stackitem.linestyle)
        self.assertIsNone(stackitem.linewidth)
        self.assertIsNone(stackitem.color)
        self.assertIsNone(stackitem.edgecolor)

    def test_style(self):
        """Check that style contains all styles"""
        stackitem = StackItem(["ggh", "vbf"], "Signal", linestyle=':',
                              color='r', linewidth=3, edgecolor='b')

        self.assertEqual(stackitem.style,
                         {'color': '#ff0000', 'edgecolor': '#0000ff',
                          'linewidth': 3, 'linestyle': ':'})

    def test_style_copy(self):
        """Check that style is a copy of the internal dict"""
        stackitem = StackItem(["ggh", "vbf"], "Signal", linestyle=':',
                              color='r', linewidth=3, edgecolor='b')

        style = stackitem.style
        style["color"] = 'xxx'

        self.assertEqual(stackitem.style["color"], "#ff0000")




import unittest
import numpy as np
from uhepp import Yield

class YieldTestCase(unittest.TestCase):
    """Test the implementation of Yield"""

    def test_zeros(self):
        """Check that zeros returns zeros of base-length"""
        data = Yield([1, 2, 3, 4])
        self.assertEqual(len(data._zeros()), 4)

    def test_sanitize_present(self):
        """Check that a array is returned if a list is given"""
        data = Yield([1, 2, 3, 4])
        array = data._sanitize([1, 2, 3, 4])
        self.assertEqual(list(array), [1, 2, 3, 4])
        self.assertIsInstance(array, np.ndarray)

    def test_sanitize_missing(self):
        """Check that a None is returned if None is given"""
        data = Yield([1, 2, 3])
        missing = data._sanitize(None)
        self.assertIsNone(missing)

    def test_sanitize_length(self):
        """Check sanitize raises a ValueEror on bin count mismatch"""
        data = Yield([1, 2, 3])
        self.assertRaises(ValueError, data._sanitize, [1, 2, 3, 4])

    def test_linear(self):
        """Check that the arrays are added linearly"""
        array_a = np.array([1, 2, 3])
        array_b = np.array([4, 5, 2])

        result = Yield._linear(array_a, array_b, lambda a, b: a + b)
        self.assertEqual(list(result), [5, 7, 5])

    def test_linear_None(self):
        """Check that the None arrays are ignored"""
        array_a = np.array([1, 2, 3])

        result = Yield._linear(array_a, None, lambda a, b: a + b)
        self.assertEqual(list(result), [1, 2, 3])

        result = Yield._linear(None, array_a, lambda a, b: a + b)
        self.assertEqual(list(result), [1, 2, 3])

        result = Yield._linear(None, None, lambda a, b: a + b)
        self.assertIsNone(result)

    def test_quadratic(self):
        """Check that the arrays are added quadratic"""
        array_a = np.array([1, 2, 3])
        array_b = np.array([4, 5, 2])

        result = Yield._quadratic(array_a, array_b, lambda a, b: a + b)
        self.assertListAlmostEqual([x**2 for x in result],
                                   [17, 29, 13])

    def test_quadratic_None(self):
        """Check that the None arrays are ignored"""
        array_a = np.array([1, 2, 3])

        result = Yield._quadratic(array_a, None, lambda a, b: a + b)
        self.assertListAlmostEqual(list(result), [1, 2, 3])

        result = Yield._quadratic(None, array_a, lambda a, b: a + b)
        self.assertListAlmostEqual(list(result), [1, 2, 3])

        result = Yield._quadratic(None, None, lambda a, b: a + b)
        self.assertIsNone(result)

    def test_optional_args(self):
        """Check that all arguments except base are optional"""
        data = Yield([1, 2, 3, 4])

        self.assertEqual(list(data._base), [1, 2, 3, 4])
        self.assertIsInstance(data._base, np.ndarray)

    def test_all_args_store(self):
        """Check that the values for all arguments are stored internally"""
        data = Yield([1, 2, 3, 4],
                     [0.1, 0.2, 0.3, 0.4],
                     [1, 1, 1, 1],
                     var_up={'exp': [1, 2, 2, 1]},
                     var_down={'theo': [2, 1, 1, 2]})

        self.assertEqual(list(data._base), [1, 2, 3, 4])
        self.assertEqual(list(data._stat), [0.1, 0.2, 0.3, 0.4])
        self.assertEqual(list(data._syst), [1, 1, 1, 1])
        self.assertIsInstance(data._base, np.ndarray)
        self.assertIsInstance(data._stat, np.ndarray)
        self.assertIsInstance(data._syst, np.ndarray)

        self.assertEqual(list(data._var_up.keys()), ['exp'])
        self.assertEqual(list(data._var_down.keys()), ['theo'])

        self.assertEqual(list(data._var_up['exp']), [1, 2, 2, 1])
        self.assertEqual(list(data._var_down['theo']), [2, 1, 1, 2])
        self.assertIsInstance(data._var_up['exp'], np.ndarray)
        self.assertIsInstance(data._var_down['theo'], np.ndarray)

    def test_base(self):
        """Check that base returns the base yield"""
        data = Yield([1, 2, 3, 4])

        self.assertEqual(data.base, [1, 2, 3, 4])
        self.assertIsInstance(data.base, list)

    def test_base_copy(self):
        """Check that base returns a copy of the base yield"""
        data = Yield([1, 2, 3, 4])

        self.assertEqual(data.base, [1, 2, 3, 4])
        base = data.base    
        base.append(5)
        self.assertEqual(data.base, [1, 2, 3, 4])

    def test_stat(self):
        """Check that stat returns the stat uncertainty"""
        data = Yield([1, 2, 3, 4],
                     [0.1, 0.2, 0.3, 0.4])

        self.assertEqual(data.stat, [0.1, 0.2, 0.3, 0.4])
        self.assertIsInstance(data.stat, list)

    def test_stat_fill(self):
        """Check that stat returns the 0 if stat is missing"""
        data = Yield([1, 2, 3, 4])

        self.assertEqual(data.stat, [0.0, 0.0, 0.0, 0.0])
        self.assertIsInstance(data.stat, list)

    def test_stat_copy(self):
        """Check that stat returns a copy of the stat uncertainty"""
        data = Yield([1, 2, 3, 4],
                     [0.1, 0.2, 0.3, 0.4])

        self.assertEqual(data.stat, [0.1, 0.2, 0.3, 0.4])
        stat = data.stat    
        stat.append(5)
        self.assertEqual(data.stat, [0.1, 0.2, 0.3, 0.4])

    def test_syst(self):
        """Check that syst returns the syst uncertainty"""
        data = Yield([1, 2, 3, 4],
                     syst=[0.1, 0.2, 0.3, 0.4])

        self.assertEqual(data.syst, [0.1, 0.2, 0.3, 0.4])
        self.assertIsInstance(data.syst, list)

    def test_syst_fill(self):
        """Check that syst returns the 0 if syst is missing"""
        data = Yield([1, 2, 3, 4])

        self.assertEqual(data.syst, [0.0, 0.0, 0.0, 0.0])
        self.assertIsInstance(data.syst, list)

    def test_syst_copy(self):
        """Check that syst returns a copy of the syst uncertainty"""
        data = Yield([1, 2, 3, 4],
                     syst=[0.1, 0.2, 0.3, 0.4])

        self.assertEqual(data.syst, [0.1, 0.2, 0.3, 0.4])
        syst = data.syst    
        syst.append(5)
        self.assertEqual(data.syst, [0.1, 0.2, 0.3, 0.4])

    def test_var_names(self):
        """Check that variation_names returns a list of all variations"""
        data = Yield([1, 2, 3, 4],
                     var_up={"tes": [2, 2, 2, 2], "theory": [2, 4, 6, 8]},
                     var_down={"jes": [2, 2, 4, 4], "theory": [2, 4, 6, 8]})

        self.assertEqual(set(data.variations),
                         set(["tes", "jes", "theory"]))
        self.assertIsInstance(data.variations, list)

        # No up
        data = Yield([1, 2, 3, 4],
                     var_down={"jes": [2, 2, 4, 4], "theory": [2, 4, 6, 8]})

        self.assertEqual(set(data.variations),
                         set(["jes", "theory"]))

        # No down
        data = Yield([1, 2, 3, 4],
                     var_up={"jes": [2, 2, 4, 4], "theory": [2, 4, 6, 8]})

        self.assertEqual(set(data.variations),
                         set(["jes", "theory"]))

    def test_var_names__no_var(self):
        """Check that variation_names returns an empty list"""
        data = Yield([1, 2, 3, 4])

        self.assertEqual(data.variations, [])

    def test_variation(self):
        """Check that variation_up/down returns the yield if found in dicts"""
        data = Yield([1, 2, 3, 4],
                     var_up={"tes": [2, 2, 2, 2], "theory": [2, 4, 6, 8]},
                     var_down={"jes": [2, 2, 1, 1], "theory": [-2, -4, -6, -8]})

        self.assertEqual(data.var_up("theory"), [2, 4, 6, 8])
        self.assertEqual(data.var_down("theory"), [-2, -4, -6, -8])

    def test_variation__onsided_this(self):
        """Test variation_up/down with one-sided var found in dict"""
        data = Yield([1, 2, 3, 4],
                     var_up={"tes": [2, 2, 2, 2], "theory": [2, 4, 6, 8]},
                     var_down={"jes": [2, 2, 1, 1], "theory": [-2, -4, -6, -8]})

        self.assertEqual(data.var_up("tes"), [2, 2, 2, 2])
        self.assertEqual(data.var_down("jes"), [2, 2, 1, 1])

        self.assertIsInstance(data.var_up("tes"), list)
        self.assertIsInstance(data.var_down("jes"), list)

    def test_variation__onsided_other(self):
        """Test variation_up/down with one-sided var found in other dict"""
        data = Yield([1, 2, 3, 4],
                     var_up={"tes": [2, 2, 2, 2], "theory": [2, 4, 6, 8]},
                     var_down={"jes": [2, 2, 1, 1], "theory": [-2, -4, -6, -8]})

        self.assertEqual(data.var_up("jes"), [0, 2, 5, 7]) 
        self.assertEqual(data.var_down("tes"), [0, 2, 4, 6])

        self.assertIsInstance(data.var_up("jes"), list)
        self.assertIsInstance(data.var_down("tes"), list)

    def test_variation_unknown(self):
        """Test variation_up/down with unknown variation"""
        data = Yield([1, 2, 3, 4],
                     var_up={"tes": [2, 2, 2, 2], "theory": [2, 4, 6, 8]},
                     var_down={"jes": [2, 2, 1, 1], "theory": [-2, -4, -6, -8]})

        self.assertEqual(data.var_up("ckkw"), [1, 2, 3, 4])
        self.assertEqual(data.var_down("ckkw"), [1, 2, 3, 4])

        self.assertIsInstance(data.var_up("ckkw"), list)
        self.assertIsInstance(data.var_down("ckkw"), list)

    def test_iter_var(self):
        """Check that iter_vars returns all names and variations"""
        data = Yield([1, 2, 3, 4],
                     var_up={"tes": [2, 2, 2, 2], "theory": [2, 4, 6, 8]},
                     var_down={"jes": [2, 2, 1, 1], "theory": [-2, -4, -6, -8]})

        items = sorted(data.iter_vars())
        self.assertEqual(items[0], ("jes", [0, 2, 5, 7], [2, 2, 1, 1]))
        self.assertEqual(items[1], ("tes", [2, 2, 2, 2],  [0, 2, 4, 6]))
        self.assertEqual(items[2], ("theory",
                                    [2, 4, 6, 8],
                                    [-2, -4, -6, -8]))
        self.assertEqual(len(items), 3)

    def test_add_var_new(self):
        """Check add_var adds a new up/down variation"""
        data = Yield([1, 2, 3, 4])

        data.add_var("tes", var_up=[2, 3, 4, 5])
        self.assertEqual(data.variations, ["tes"])
        self.assertEqual(data.var_up("tes"), [2, 3, 4, 5])

        data.add_var("jes", var_down=[2, 3, 4, 5])
        self.assertEqual(sorted(data.variations), ["jes", "tes"])
        self.assertEqual(data.var_down("jes"), [2, 3, 4, 5])

        data.add_var("theory", var_up=[1, 1, 1, 1], var_down=[3, 3, 3, 3])
        self.assertEqual(sorted(data.variations),
                          ["jes", "tes", "theory"])
        self.assertEqual(data.var_up("theory"), [1, 1, 1, 1])
        self.assertEqual(data.var_down("theory"), [3, 3, 3, 3])

    def test_add_var_overwrite(self):
        """Check add_var overwrites existing up/down variations"""
        data = Yield([1, 2, 3, 4],
                     var_up={"tes": [2, 2, 2, 2], "theory": [2, 4, 6, 8]},
                     var_down={"jes": [2, 2, 1, 1], "theory": [-2, -4, -6, -8]})

        data.add_var("tes", var_up=[2, 3, 4, 5])
        self.assertEqual(data.var_up("tes"), [2, 3, 4, 5])

        data.add_var("jes", var_down=[2, 3, 4, 5])
        self.assertEqual(data.var_down("jes"), [2, 3, 4, 5])

        data.add_var("theory", var_up=[1, 1, 1, 1], var_down=[3, 3, 3, 3])
        self.assertEqual(data.var_up("theory"), [1, 1, 1, 1])
        self.assertEqual(data.var_down("theory"), [3, 3, 3, 3])


    def assertListAlmostEqual(self, iter_a, iter_b, *args, **kwds):
        """Check that the contents of both lists are almost equal"""
        for item_a, item_b in zip(iter_a, iter_b):
            self.assertAlmostEqual(item_a, item_b, *args, **kwds)

        self.assertEqual(len(iter_a), len(iter_b))

    def test_add_number(self):
        """Check the result of adding a number to a yield"""
        data = Yield([1, 2, 3, 4], [0.1, 0.2, 0.3, 0.4], [0.2, 0.2, 0.4, 0.4],
                     var_up={"tes": [2, 2, 2, 2], "theory": [2, 4, 6, 8]},
                     var_down={"jes": [2, 2, 1, 1], "theory": [-2, -4, -6, -8]})

        data = data + 5
        self.assertEqual(data.base, [6, 7, 8, 9])
        self.assertEqual(data.stat, [0.1, 0.2, 0.3, 0.4])
        self.assertEqual(data.syst, [0.2, 0.2, 0.4, 0.4])

        self.assertEqual(data.var_up("tes"), [7, 7, 7, 7])
        self.assertEqual(data.var_down("tes"), [5, 7, 9, 11])

        self.assertEqual(data.var_up("jes"), [5, 7, 10, 12])
        self.assertEqual(data.var_down("jes"), [7, 7, 6, 6])

        self.assertEqual(data.var_up("theory"), [7, 9, 11, 13])
        self.assertEqual(data.var_down("theory"), [3, 1, -1, -3])

    def test_add_yield(self):
        """Check the result of adding two yields"""
        data_a = Yield([1, 2, 3, 4], [0.1, 0.2, 0.3, 0.4], [0.2, 0.2, 0.4, 0.4],
                       var_up={"tes": [2, 2, 2, 2],
                               "theory": [2, 4, 6, 8]},
                       var_down={"jes": [2, 2, 1, 1],
                                 "theory": [-2, -4, -6, -8]})

        data_b = Yield([0, 3, 2, 1], [0.1, 0.2, 0.2, 0.1], [0.2, 0.3, 0.3, 0.3],
                       var_up={"jer": [1, 2, 2, 1],
                               "theory": [2, 4, 6, 8]},
                       var_down={"jes": [1, 1, 1, 1],
                                 "theory": [-1, -2, -3, -4]})

        data = data_a + data_b
        self.assertEqual(data.base, [1, 5, 5, 5])
        self.assertListAlmostEqual([x**2 for x in data.stat], [0.02, 0.08, 0.13, 0.17])
        self.assertListAlmostEqual([x**2 for x in data.syst], [0.08, 0.13, 0.25, 0.25])

        self.assertEqual(data.var_up("tes"), [2, 5, 4, 3])
        self.assertEqual(data.var_down("tes"), [0, 5, 6, 7])

        self.assertEqual(data.var_up("jes"), [-1, 7, 8, 8])
        self.assertEqual(data.var_down("jes"), [3, 3, 2, 2])

        self.assertEqual(data.var_up("jer"), [2, 4, 5, 5])
        self.assertEqual(data.var_down("jer"), [0, 6, 5, 5])

        self.assertEqual(data.var_up("theory"), [4, 8, 12, 16])
        self.assertEqual(data.var_down("theory"), [-3, -6, -9, -12])

    def test_add_no_uncert_self(self):
        """Check adding yields without self uncertainty"""
        data_a = Yield([1, 2, 3, 4])

        data_b = Yield([0, 3, 2, 1], [0.1, 0.2, 0.2, 0.1], [0.2, 0.3, 0.3, 0.3],
                       var_up={"jer": [1, 2, 2, 1],
                               "theory": [2, 4, 6, 8]},
                       var_down={"jes": [1, 1, 1, 1],
                                 "theory": [-1, -2, -3, -4]})

        data = data_a + data_b
        self.assertEqual(data.base, [1, 5, 5, 5])
        self.assertListAlmostEqual(data.stat, [0.1, 0.2, 0.2, 0.1])
        self.assertListAlmostEqual(data.syst, [0.2, 0.3, 0.3, 0.3])

        self.assertEqual(data.var_up("jes"), [0, 7, 6, 5])
        self.assertEqual(data.var_down("jes"), [2, 3, 4, 5])

        self.assertEqual(data.var_up("jer"), [2, 4, 5, 5])
        self.assertEqual(data.var_down("jer"), [0, 6, 5, 5])

        self.assertEqual(data.var_up("theory"), [3, 6, 9, 12])
        self.assertEqual(data.var_down("theory"), [0, 0, 0, 0])

    def test_add_no_uncert_other(self):
        data_a = Yield([1, 2, 3, 4])

        data_b = Yield([0, 3, 2, 1], [0.1, 0.2, 0.2, 0.1], [0.2, 0.3, 0.3, 0.3],
                       var_up={"jer": [1, 2, 2, 1],
                               "theory": [2, 4, 6, 8]},
                       var_down={"jes": [1, 1, 1, 1],
                                 "theory": [-1, -2, -3, -4]})

        data = data_b + data_a
        self.assertEqual(data.base, [1, 5, 5, 5])
        self.assertListAlmostEqual(data.stat, [0.1, 0.2, 0.2, 0.1])
        self.assertListAlmostEqual(data.syst, [0.2, 0.3, 0.3, 0.3])

        self.assertEqual(data.var_up("jes"), [0, 7, 6, 5])
        self.assertEqual(data.var_down("jes"), [2, 3, 4, 5])

        self.assertEqual(data.var_up("jer"), [2, 4, 5, 5])
        self.assertEqual(data.var_down("jer"), [0, 6, 5, 5])

        self.assertEqual(data.var_up("theory"), [3, 6, 9, 12])
        self.assertEqual(data.var_down("theory"), [0, 0, 0, 0])

    def test_add_different_bin_count(self):
        """Check that a value error is raised if the bin count differs"""
        data_a = Yield([1, 2, 3, 4])
        data_b = Yield([0, 3, 2, 1, 5])

        self.assertRaises(ValueError, lambda: data_a + data_b)

    def test_copy_init_args(self):
        """Check that init argument arrays are copied"""
        array = np.array([1, 2, 3, 4])
        data = Yield(array, array, array,
                     var_up={"jer": array},
                     var_down={"jes": array})

        array[0] = -1
        self.assertEqual(data.base, [1, 2, 3, 4])
        self.assertEqual(data.stat, [1, 2, 3, 4])
        self.assertEqual(data.syst, [1, 2, 3, 4])
        self.assertEqual(data.var_up("jer"), [1, 2, 3, 4])
        self.assertEqual(data.var_down("jes"), [1, 2, 3, 4])

    def test_init_array_length(self):
        """Check that calling init with inconsistent bins raises an error"""
        self.assertRaises(ValueError, Yield, [1, 2, 3, 4], [1, 2, 3])
        self.assertRaises(ValueError, Yield, [1, 2, 3, 4], syst=[1, 2, 3])
        self.assertRaises(ValueError, Yield, [1, 2, 3, 4],
                          var_up={"tes": [1, 2, 3]})
        self.assertRaises(ValueError, Yield, [1, 2, 3, 4],
                          var_down={"tes": [1, 2, 3]})

    def test_var_add(self):
        """Check that adding a var with inconsistent bins raises an error"""
        data = Yield([1, 2, 3, 4])
        self.assertRaises(ValueError, data.add_var, "tes", var_up=[1, 2, 3])
        self.assertRaises(ValueError, data.add_var, "tes", var_down=[1, 2, 3])

    def test_radd_number(self):
        """Check the result of adding a number to a yield"""
        data = Yield([1, 2, 3, 4], [0.1, 0.2, 0.3, 0.4], [0.2, 0.2, 0.4, 0.4],
                     var_up={"tes": [2, 2, 2, 2], "theory": [2, 4, 6, 8]},
                     var_down={"jes": [2, 2, 1, 1], "theory": [-2, -4, -6, -8]})

        data = 5 + data
        self.assertEqual(data.base, [6, 7, 8, 9])
        self.assertEqual(data.stat, [0.1, 0.2, 0.3, 0.4])
        self.assertEqual(data.syst, [0.2, 0.2, 0.4, 0.4])

        self.assertEqual(data.var_up("tes"), [7, 7, 7, 7])
        self.assertEqual(data.var_down("tes"), [5, 7, 9, 11])

        self.assertEqual(data.var_up("jes"), [5, 7, 10, 12])
        self.assertEqual(data.var_down("jes"), [7, 7, 6, 6])

        self.assertEqual(data.var_up("theory"), [7, 9, 11, 13])
        self.assertEqual(data.var_down("theory"), [3, 1, -1, -3])

    ####### Subtract
    def test_sub_number(self):
        """Check the result of subtracting a number to a yield"""
        data = Yield([1, 2, 3, 4], [0.1, 0.2, 0.3, 0.4], [0.2, 0.2, 0.4, 0.4],
                     var_up={"tes": [2, 2, 2, 2], "theory": [2, 4, 6, 8]},
                     var_down={"jes": [2, 2, 1, 1], "theory": [-2, -4, -6, -8]})

        data = data - (-5)
        self.assertEqual(data.base, [6, 7, 8, 9])
        self.assertEqual(data.stat, [0.1, 0.2, 0.3, 0.4])
        self.assertEqual(data.syst, [0.2, 0.2, 0.4, 0.4])

        self.assertEqual(data.var_up("tes"), [7, 7, 7, 7])
        self.assertEqual(data.var_down("tes"), [5, 7, 9, 11])

        self.assertEqual(data.var_up("jes"), [5, 7, 10, 12])
        self.assertEqual(data.var_down("jes"), [7, 7, 6, 6])

        self.assertEqual(data.var_up("theory"), [7, 9, 11, 13])
        self.assertEqual(data.var_down("theory"), [3, 1, -1, -3])

    def test_sub_yield(self):
        """Check the result of subing two yields"""
        data_a = Yield([1, 2, 3, 4], [0.1, 0.2, 0.3, 0.4], [0.2, 0.2, 0.4, 0.4],
                       var_up={"tes": [2, 2, 2, 2],
                               "theory": [2, 4, 6, 8]},
                       var_down={"jes": [2, 2, 1, 1],
                                 "theory": [-2, -4, -6, -8]})

        data_b = Yield([-0, -3, -2, -1], [0.1, 0.2, 0.2, 0.1], [0.2, 0.3, 0.3, 0.3],
                       var_up={"jer": [-1, -2, -2, -1],
                               "theory": [-2, -4, -6, -8]},
                       var_down={"jes": [-1, -1, -1, -1],
                                 "theory": [1, 2, 3, 4]})

        data = data_a - data_b
        self.assertEqual(data.base, [1, 5, 5, 5])
        self.assertListAlmostEqual([x**2 for x in data.stat], [0.02, 0.08, 0.13, 0.17])
        self.assertListAlmostEqual([x**2 for x in data.syst], [0.08, 0.13, 0.25, 0.25])

        self.assertEqual(data.var_up("tes"), [2, 5, 4, 3])
        self.assertEqual(data.var_down("tes"), [0, 5, 6, 7])

        self.assertEqual(data.var_up("jes"), [-1, 7, 8, 8])
        self.assertEqual(data.var_down("jes"), [3, 3, 2, 2])

        self.assertEqual(data.var_up("jer"), [2, 4, 5, 5])
        self.assertEqual(data.var_down("jer"), [0, 6, 5, 5])

        self.assertEqual(data.var_up("theory"), [4, 8, 12, 16])
        self.assertEqual(data.var_down("theory"), [-3, -6, -9, -12])

    def test_sub_no_uncert_self(self):
        """Check subing yields without self uncertainty"""
        data_a = Yield([1, 2, 3, 4])

        data_b = Yield([-0, -3, -2, -1], [0.1, 0.2, 0.2, 0.1], [0.2, 0.3, 0.3, 0.3],
                       var_up={"jer": [-1, -2, -2, -1],
                               "theory": [-2, -4, -6, -8]},
                       var_down={"jes": [-1, -1, -1, -1],
                                 "theory": [1, 2, 3, 4]})

        data = data_a - data_b
        self.assertEqual(data.base, [1, 5, 5, 5])
        self.assertListAlmostEqual(data.stat, [0.1, 0.2, 0.2, 0.1])
        self.assertListAlmostEqual(data.syst, [0.2, 0.3, 0.3, 0.3])

        self.assertEqual(data.var_up("jes"), [0, 7, 6, 5])
        self.assertEqual(data.var_down("jes"), [2, 3, 4, 5])

        self.assertEqual(data.var_up("jer"), [2, 4, 5, 5])
        self.assertEqual(data.var_down("jer"), [0, 6, 5, 5])

        self.assertEqual(data.var_up("theory"), [3, 6, 9, 12])
        self.assertEqual(data.var_down("theory"), [0, 0, 0, 0])

    def test_sub_no_uncert_other(self):
        data_a = Yield([-1, -2, -3, -4])

        data_b = Yield([0, 3, 2, 1], [0.1, 0.2, 0.2, 0.1], [0.2, 0.3, 0.3, 0.3],
                       var_up={"jer": [1, 2, 2, 1],
                               "theory": [2, 4, 6, 8]},
                       var_down={"jes": [1, 1, 1, 1],
                                 "theory": [-1, -2, -3, -4]})

        data = data_b - data_a
        self.assertEqual(data.base, [1, 5, 5, 5])
        self.assertListAlmostEqual(data.stat, [0.1, 0.2, 0.2, 0.1])
        self.assertListAlmostEqual(data.syst, [0.2, 0.3, 0.3, 0.3])

        self.assertEqual(data.var_up("jes"), [0, 7, 6, 5])
        self.assertEqual(data.var_down("jes"), [2, 3, 4, 5])

        self.assertEqual(data.var_up("jer"), [2, 4, 5, 5])
        self.assertEqual(data.var_down("jer"), [0, 6, 5, 5])

        self.assertEqual(data.var_up("theory"), [3, 6, 9, 12])
        self.assertEqual(data.var_down("theory"), [0, 0, 0, 0])

    def test_sub_different_bin_count(self):
        """Check that a value error is raised if the bin count differs"""
        data_a = Yield([1, 2, 3, 4])
        data_b = Yield([0, 3, 2, 1, 5])

        self.assertRaises(ValueError, lambda: data_a - data_b)

    def test_rsub_number(self):
        """Check the result of subtracting a number to a yield"""
        data = Yield([1, 2, 3, 4], [0.1, 0.2, 0.3, 0.4], [0.2, 0.2, 0.4, 0.4],
                     var_up={"tes": [2, 2, 2, 2], "theory": [2, 4, 6, 8]},
                     var_down={"jes": [2, 2, 1, 1], "theory": [-2, -4, -6, -8]})

        data = -5 - data
        self.assertEqual(data.base, [-6, -7, -8, -9])
        self.assertEqual(data.stat, [0.1, 0.2, 0.3, 0.4])
        self.assertEqual(data.syst, [0.2, 0.2, 0.4, 0.4])

        self.assertEqual(data.var_up("tes"), [-7, -7, -7, -7])
        self.assertEqual(data.var_down("tes"), [-5, -7, -9, -11])

        self.assertEqual(data.var_up("jes"), [-5, -7, -10, -12])
        self.assertEqual(data.var_down("jes"), [-7, -7, -6, -6])

        self.assertEqual(data.var_up("theory"), [-7, -9, -11, -13])
        self.assertEqual(data.var_down("theory"), [-3, -1, 1, 3])

    ####### Neg
    def test_neg(self):
        """Check that -Yield inverts base and variations"""
        data = Yield([0, 3, 2, 1], [0.1, 0.2, 0.2, 0.1], [0.2, 0.3, 0.3, 0.3],
                     var_up={"jer": [1, 2, 2, 1], "theory": [-2, -4, -6, -8]},
                     var_down={"jes": [1, 1, 1, 1], "theory": [1, 2, 3, 4]})

        data = -data

        self.assertEqual(data.base, [-0, -3, -2, -1])
        self.assertEqual(data.stat, [0.1, 0.2, 0.2, 0.1])
        self.assertEqual(data.syst, [0.2, 0.3, 0.3, 0.3])
        self.assertEqual(data.var_up("jer"), [-1, -2, -2, -1])
        self.assertEqual(data.var_up("theory"), [2, 4, 6, 8])
        self.assertEqual(data.var_down("jes"), [-1, -1, -1, -1])
        self.assertEqual(data.var_down("theory"), [-1, -2, -3, -4])

    ####### Pos
    def test_pos(self):
        """Check that +Yield returns base and variations"""
        data = Yield([0, 3, 2, 1], [0.1, 0.2, 0.2, 0.1], [0.2, 0.3, 0.3, 0.3],
                     var_up={"jer": [1, 2, 2, 1], "theory": [-2, -4, -6, -8]},
                     var_down={"jes": [1, 1, 1, 1], "theory": [1, 2, 3, 4]})

        data = +data

        self.assertEqual(data.base, [0, 3, 2, 1])
        self.assertEqual(data.stat, [0.1, 0.2, 0.2, 0.1])
        self.assertEqual(data.syst, [0.2, 0.3, 0.3, 0.3])
        self.assertEqual(data.var_up("jer"), [1, 2, 2, 1])
        self.assertEqual(data.var_up("theory"), [-2, -4, -6, -8])
        self.assertEqual(data.var_down("jes"), [1, 1, 1, 1])
        self.assertEqual(data.var_down("theory"), [1, 2, 3, 4])

    ####### Len
    def test_len(self):
        """Check that len returns the length of the base"""
        data = Yield([0, 3, 2, 1])
        self.assertEqual(len(data), 4)

    ####### getitem
    def test_getitem(self):
        """Check that getitem returns an element of base"""
        data = Yield([0, 3, 2, 1])

        self.assertEqual(data[0], 0)
        self.assertEqual(data[1], 3)
        self.assertEqual(data[2], 2)
        self.assertEqual(data[3], 1)

    def test_getitem_index(self):
        """Check that an Error is raised if the index is invalid"""
        data = Yield([0, 3, 2, 1])
        self.assertRaises(IndexError, lambda: data[4])

    def test_getitem_convert(self):
        """Check that a yield object can be iterated"""
        data = Yield([0, 3, 2, 1])

        self.assertEqual(list(data), [0, 3, 2, 1])

    ####### mul
    def test_mul_number(self):
        """Check the result of multiplying a number with a yield"""
        data = Yield([1, 2, 3, 4], [0.1, 0.2, 0.3, 0.4], [0.2, 0.2, 0.4, 0.4],
                     var_up={"tes": [2, 2, 2, 2], "theory": [2, 4, 6, 8]},
                     var_down={"jes": [2, 2, 1, 1], "theory": [-2, -4, -6, -8]})

        data = data * (-2)
        self.assertEqual(data.base, [-2, -4, -6, -8])
        self.assertEqual(data.stat, [0.2, 0.4, 0.6, 0.8])
        self.assertEqual(data.syst, [0.4, 0.4, 0.8, 0.8])

        self.assertEqual(data.var_up("tes"), [-4, -4, -4, -4])
        self.assertEqual(data.var_down("tes"), [0, -4, -8, -12])

        self.assertEqual(data.var_up("jes"), [-0, -4, -10, -14])
        self.assertEqual(data.var_down("jes"), [-4, -4, -2, -2])

        self.assertEqual(data.var_up("theory"), [-4, -8, -12, -16])
        self.assertEqual(data.var_down("theory"), [4, 8, 12, 16])

    def test_mul_number_no_unert(self):
        """Check the result of multiplying a number with a yield"""
        data = Yield([1, 2, 3, 4])

        data = data * (-2)
        self.assertEqual(data.base, [-2, -4, -6, -8])
        self.assertIsNone(data._stat)
        self.assertIsNone(data._syst)

    def test_mul_yield(self):
        """Check the result of multiplying two yields"""
        data_a = Yield([1, 2, 3, 4], [0.1, 0.2, 0.3, 0.4], [0.2, 0.2, 0.4, 0.4],
                       var_up={"tes": [2, 2, 2, 2],
                               "theory": [2, 4, 6, 8]},
                       var_down={"jes": [2, 2, 1, 1],
                                 "theory": [-2, -4, -6, -8]})

        data_b = Yield([0, 3, 2, 1], [0.1, 0.2, 0.2, 0.1], [0.2, 0.3, 0.3, 0.3],
                       var_up={"jer": [1, 2, 2, 1],
                               "theory": [2, 4, 6, 8]},
                       var_down={"jes": [1, 1, 1, 1],
                                 "theory": [-1, -2, -3, -4]})

        data = data_a * data_b
        self.assertEqual(data.base, [0, 6, 6, 4])
        self.assertListAlmostEqual([x**2 for x in data.stat],
                                   [0.01, 0.52, 0.72, 0.32])
        self.assertListAlmostEqual([x**2 for x in data.syst],
                                   [0.04, 0.72, 1.45, 1.60])

        self.assertEqual(data.var_up("tes"), [0, 6, 4, 2])
        self.assertEqual(data.var_down("tes"), [0, 6, 8, 6])

        self.assertEqual(data.var_up("jes"), [0, 10, 15, 7])
        self.assertEqual(data.var_down("jes"), [2, 2, 1, 1])

        self.assertEqual(data.var_up("jer"), [1, 4, 6, 4])
        self.assertEqual(data.var_down("jer"), [-1, 8, 6, 4])

        self.assertEqual(data.var_up("theory"), [4, 16, 36, 64])
        self.assertEqual(data.var_down("theory"), [2, 8, 18, 32])

    def test_mul_no_uncert_self(self):
        """Check muling yields without self uncertainty"""
        data_a = Yield([1, 2, 3, 4])

        data_b = Yield([0, 3, 2, 1], [0.1, 0.2, 0.2, 0.1], [0.2, 0.3, 0.3, 0.3],
                       var_up={"jer": [1, 2, 2, 1],
                               "theory": [2, 4, 6, 8]},
                       var_down={"jes": [1, 1, 1, 1],
                                 "theory": [-1, -2, -3, -4]})

        data = data_a * data_b
        self.assertEqual(data.base, [0, 6, 6, 4])
        self.assertListAlmostEqual(data.stat, [0.1, 0.4, 0.6, 0.4])
        self.assertListAlmostEqual(data.syst, [0.2, 0.6, 0.9, 1.2])

        self.assertEqual(data.var_up("jes"), [-1, 10, 9, 4])
        self.assertEqual(data.var_down("jes"), [1, 2, 3, 4])

        self.assertEqual(data.var_up("jer"), [1, 4, 6, 4])
        self.assertEqual(data.var_down("jer"), [-1, 8, 6, 4])

        self.assertEqual(data.var_up("theory"), [2, 8, 18, 32])
        self.assertEqual(data.var_down("theory"), [-1, -4, -9, -16])

    def test_mul_no_uncert_other(self):
        data_a = Yield([1, 2, 3, 4])

        data_b = Yield([0, 3, 2, 1], [0.1, 0.2, 0.2, 0.1], [0.2, 0.3, 0.3, 0.3],
                       var_up={"jer": [1, 2, 2, 1],
                               "theory": [2, 4, 6, 8]},
                       var_down={"jes": [1, 1, 1, 1],
                                 "theory": [-1, -2, -3, -4]})

        data = data_b * data_a
        self.assertEqual(data.base, [0, 6, 6, 4])
        self.assertListAlmostEqual(data.stat, [0.1, 0.4, 0.6, 0.4])
        self.assertListAlmostEqual(data.syst, [0.2, 0.6, 0.9, 1.2])

        self.assertEqual(data.var_up("jes"), [-1, 10, 9, 4])
        self.assertEqual(data.var_down("jes"), [1, 2, 3, 4])

        self.assertEqual(data.var_up("jer"), [1, 4, 6, 4])
        self.assertEqual(data.var_down("jer"), [-1, 8, 6, 4])

        self.assertEqual(data.var_up("theory"), [2, 8, 18, 32])
        self.assertEqual(data.var_down("theory"), [-1, -4, -9, -16])

    def test_mul_different_bin_count(self):
        """Check that a value error is raised if the bin count differs"""
        data_a = Yield([1, 2, 3, 4])
        data_b = Yield([0, 3, 2, 1, 5])

        self.assertRaises(ValueError, lambda: data_a * data_b)

    ####### rmul
    def test_rmul_number(self):
        """Check the result of multiplying a number with a yield"""
        data = Yield([1, 2, 3, 4], [0.1, 0.2, 0.3, 0.4], [0.2, 0.2, 0.4, 0.4],
                     var_up={"tes": [2, 2, 2, 2], "theory": [2, 4, 6, 8]},
                     var_down={"jes": [2, 2, 1, 1], "theory": [-2, -4, -6, -8]})

        data = -2 * data
        self.assertEqual(data.base, [-2, -4, -6, -8])
        self.assertEqual(data.stat, [0.2, 0.4, 0.6, 0.8])
        self.assertEqual(data.syst, [0.4, 0.4, 0.8, 0.8])

        self.assertEqual(data.var_up("tes"), [-4, -4, -4, -4])
        self.assertEqual(data.var_down("tes"), [0, -4, -8, -12])

        self.assertEqual(data.var_up("jes"), [-0, -4, -10, -14])
        self.assertEqual(data.var_down("jes"), [-4, -4, -2, -2])

        self.assertEqual(data.var_up("theory"), [-4, -8, -12, -16])
        self.assertEqual(data.var_down("theory"), [4, 8, 12, 16])

    ####### rebin
    def test_rebin(self):
        """Check the rebinning a yield sums adjacent bins"""
        data = Yield([1, 2, 3, 4], [0.1, 0.2, 0.3, 0.4], [0.2, 0.2, 0.4, 0.4],
                     var_up={"tes": [2, 2, 2, 2], "theory": [2, 4, 6, 8]},
                     var_down={"jes": [2, 2, 1, 1], "theory": [-2, -4, -6, -8]})

        data = data.rebin([0, 1, 2], [0, 2])
        self.assertEqual(data.base, [1, 5, 4])
        self.assertListAlmostEqual([x**2 for x in data.stat], [0.01, 0.13, 0.16])
        self.assertListAlmostEqual([x**2 for x in data.syst], [0.04, 0.20, 0.16])

        self.assertEqual(data.var_up("tes"), [2, 4, 2])
        self.assertEqual(data.var_down("tes"), [0, 6, 6])

        self.assertEqual(data.var_up("jes"), [0, 7, 7])
        self.assertEqual(data.var_down("jes"), [2, 3, 1])

        self.assertEqual(data.var_up("theory"), [2, 10, 8])
        self.assertEqual(data.var_down("theory"), [-2, -10, -8])

    ####### truediv
    def test_truediv_number(self):
        """Check the result of dividing a number with a yield"""
        data = Yield([1, 2, 3, 4], [0.1, 0.2, 0.3, 0.4], [0.2, 0.2, 0.4, 0.4],
                     var_up={"tes": [2, 2, 2, 2], "theory": [2, 4, 6, 8]},
                     var_down={"jes": [2, 2, 1, 1], "theory": [-2, -4, -6, -8]})

        data = data / (-2)
        self.assertEqual(data.base, [-0.5, -1, -1.5, -2])
        self.assertEqual(data.stat, [0.05, 0.1, 0.15, 0.2])
        self.assertEqual(data.syst, [0.1, 0.1, 0.2, 0.2])

        self.assertEqual(data.var_up("tes"), [-1, -1, -1, -1])
        self.assertEqual(data.var_down("tes"), [0, -1, -2, -3])

        self.assertEqual(data.var_up("jes"), [0, -1, -2.5, -3.5])
        self.assertEqual(data.var_down("jes"), [-1, -1, -0.5, -0.5])

        self.assertEqual(data.var_up("theory"), [-1, -2, -3, -4])
        self.assertEqual(data.var_down("theory"), [1, 2, 3, 4])

    def test_truediv_number_no_unert(self):
        """Check the result of multiplying a number with a yield"""
        data = Yield([1, 2, 3, 4])

        data = data / (-2)
        self.assertEqual(data.base, [-0.5, -1, -1.5, -2])
        self.assertIsNone(data._stat)
        self.assertIsNone(data._syst)

    def test_truediv_yield(self):
        """Check the result of dividing two yields"""
        data_a = Yield([1, 2, 3, 4], [0.1, 0.2, 0.3, 0.4], [0.2, 0.2, 0.4, 0.4],
                       var_up={"tes": [2, 2, 2, 2],
                               "theory": [2, 4, 6, 8]},
                       var_down={"jes": [2, 2, 1, 1],
                                 "theory": [-2, -4, -6, -8]})

        data_b = Yield([1, 3, 2, 1], [0.1, 0.2, 0.2, 0.1], [0.2, 0.3, 0.3, 0.3],
                       var_up={"jer": [1, 2, 2, 1],
                               "theory": [2, 4, 6, 8]},
                       var_down={"jes": [1, 1, 1, 1],
                                 "theory": [-1, -2, -3, -4]})

        data = data_a / data_b
        self.assertListAlmostEqual(data.base, [1, 2/3, 1.5, 4])
        self.assertListAlmostEqual([x**2 for x in data.stat],
                                   [0.02, 0.52/81, 0.045, 0.32]) 
        self.assertListAlmostEqual([x**2 for x in data.syst],
                                   [0.08, 0.08/9, 1.45/16, 1.6])

        self.assertEqual(data.var_up("tes"), [2, 2/3, 1, 2])
        self.assertEqual(data.var_down("tes"), [0, 2/3, 2, 6])

        self.assertEqual(data.var_up("jes"), [0, 0.4, 5/3, 7])
        self.assertEqual(data.var_down("jes"), [2, 2, 1, 1])

        self.assertEqual(data.var_up("jer"), [1, 1, 3/2, 4])
        self.assertEqual(data.var_down("jer"), [1, 0.5, 3/2, 4])

        self.assertEqual(data.var_up("theory"), [1, 1, 1, 1])
        self.assertEqual(data.var_down("theory"), [2, 2, 2, 2])

    def test_truediv_no_uncert_self(self):
        """Check divison yields without self uncertainty"""
        data_a = Yield([1, 2, 3, 4])

        data_b = Yield([1, 3, 2, 1], [0.1, 0.2, 0.2, 0.1], [0.2, 0.3, 0.3, 0.3],
                       var_up={"jer": [1, 2, 2, 1],
                               "theory": [2, 4, 6, 8]},
                       var_down={"jes": [1, 1, 1, 1],
                                 "theory": [-1, -2, -3, -4]})

        data = data_a / data_b
        self.assertEqual(data.base, [1, 2/3, 1.5, 4])
        self.assertListAlmostEqual([x**2 for x in data.stat],
                                   [0.01, 0.16/81, 0.09/4, 0.16])
        self.assertListAlmostEqual([x**2 for x in data.syst],
                                   [0.04, 0.04/9, 0.81/16, 1.44])

        self.assertEqual(data.var_up("jes"), [1, 0.4, 1, 4])
        self.assertEqual(data.var_down("jes"), [1, 2, 3, 4])

        self.assertEqual(data.var_up("jer"), [1, 1, 1.5, 4])
        self.assertEqual(data.var_down("jer"), [1, 0.5, 1.5, 4])

        self.assertEqual(data.var_up("theory"), [0.5, 0.5, 0.5, 0.5])
        self.assertEqual(data.var_down("theory"), [-1, -1, -1, -1])

    def test_truediv_no_uncert_other(self):
        """Check division without other uncertainties"""
        data_b = Yield([1, 3, 2, 1], [0.1, 0.2, 0.2, 0.1], [0.2, 0.3, 0.3, 0.3],
                       var_up={"jer": [1, 2, 2, 1],
                               "theory": [2, 4, 6, 8]},
                       var_down={"jes": [1, 1, 1, 1],
                                 "theory": [-1, -2, -3, -4]})

        data_a = Yield([1, 2, 3, 4])

        data = data_b / data_a
        self.assertEqual(data.base, [1, 3/2, 2/3, 0.25])
        self.assertListAlmostEqual(data.stat, [0.1, 0.1, 0.2/3, 0.025])
        self.assertListAlmostEqual(data.syst, [0.2, 0.3/2, 0.1, 0.3/4])

        self.assertEqual(data.var_up("jes"), [1, 2.5, 1, 0.25])
        self.assertEqual(data.var_down("jes"), [1, 0.5, 1/3, 0.25])

        self.assertEqual(data.var_up("jer"), [1, 1, 2/3, 0.25])
        self.assertEqual(data.var_down("jer"), [1, 2, 2/3, 0.25])

        self.assertEqual(data.var_up("theory"), [2, 2, 2, 2])
        self.assertEqual(data.var_down("theory"), [-1, -1, -1, -1])

    def test_truediv_different_bin_count(self):
        """Check that a value error is raised if the bin count differs"""
        data_a = Yield([1, 2, 3, 4])
        data_b = Yield([0, 3, 2, 1, 5])

        self.assertRaises(ValueError, lambda: data_a / data_b)

    def test_rtruediv_number_no_unert(self):
        """Check the result of multiplying a number with a yield"""
        data = Yield([1, 2, 3, 4])

        data = -2 / data
        self.assertEqual(data.base, [-2, -1, -2/3, -0.5])
        self.assertIsNone(data._stat)
        self.assertIsNone(data._syst)


    ####### total
    def test_total(self):
        """Check the total contains all values"""
        data = Yield([1, 2, 3, 4], [0.1, 0.2, 0.3, 0.4], [0.2, 0.2, 0.4, 0.4],
                     var_up={"tes": [2, 2, 2, 2], "theory": [2, 4, 6, 8]},
                     var_down={"jes": [2, 2, 1, 1], "theory": [-2, -4, -6, -8]})

        total = data.total()

        self.assertEqual(total.base[0], 10)
        self.assertAlmostEqual(total.stat[0]**2, 0.01+0.04+0.09+0.16)
        self.assertAlmostEqual(total.syst[0]**2, 0.04+0.04+0.16+0.16)

        self.assertEqual(total.var_up("jes"), [14])
        self.assertEqual(total.var_down("jes"), [6])

        self.assertEqual(total.var_up("tes"), [8])
        self.assertEqual(total.var_down("tes"), [12])

        self.assertEqual(total.var_up("theory"), [20])
        self.assertEqual(total.var_down("theory"), [-20])

    def test_sum(self):
        """Check the total contains all values"""
        data = Yield([1, 2, 3, 4], [0.1, 0.2, 0.3, 0.4], [0.2, 0.2, 0.4, 0.4],
                     var_up={"tes": [2, 2, 2, 2], "theory": [2, 4, 6, 8]},
                     var_down={"jes": [2, 2, 1, 1], "theory": [-2, -4, -6, -8]})

        total = data.sum()

        self.assertEqual(total, 10)


    ######## Vary
    def test_vary_simple(self):
        """Check that vary return up, down or base"""
        data = Yield([1, 2, 3, 4],
                     var_up={"var1": [1.1, 2, 3, 4],
                             "var2": [1, 2.1, 3, 4],
                             "var3": [1, 2.1, 3.1, 4]},
                     var_down={"var1": [0.8, 2, 3, 4],
                               "var2": [1, 1.8, 3, 4],
                               "var4": [1, 2, 3.2, 4.2]})

        self.assertListAlmostEqual(data.vary(), [1, 2, 3, 4])
        self.assertListAlmostEqual(data.vary(var1=1), [1.1, 2, 3, 4])
        self.assertListAlmostEqual(data.vary(var2=0), [1, 2, 3, 4])
        self.assertListAlmostEqual(data.vary(var3=-1), [1, 1.9, 2.9, 4])
        self.assertListAlmostEqual(data.vary(var4=1), [1, 2, 2.8, 3.8])
            
    def test_vary_interpolate(self):
        """Check that vary interpolates between -1 and 1"""
        data = Yield([1, 2, 3, 4],
                     var_up={"var1": [1.1, 2, 3, 4],
                             "var2": [1, 2.1, 3, 4],
                             "var3": [1, 2.1, 3.1, 4]},
                     var_down={"var1": [0.8, 2, 3, 4],
                               "var2": [1, 1.8, 3, 4],
                               "var4": [1, 2, 3.2, 4.2]})

        self.assertListAlmostEqual(data.vary(var1=0.5), [1.05, 2, 3, 4])
        self.assertListAlmostEqual(data.vary(var2=0.5), [1, 2.05, 3, 4])
        self.assertListAlmostEqual(data.vary(var3=-0.5), [1, 1.95, 2.95, 4])
        self.assertListAlmostEqual(data.vary(var4=0.5), [1, 2, 2.9, 3.9])
            
    def test_vary_extrapolate(self):
        """Check that vary extrapolate outside -1 and 1"""
        data = Yield([1, 2, 3, 4],
                     var_up={"var1": [1.1, 2, 3, 4],
                             "var2": [1, 2.1, 3, 4],
                             "var3": [1, 2.1, 3.1, 4]},
                     var_down={"var1": [0.8, 2, 3, 4],
                               "var2": [1, 1.8, 3, 4],
                               "var4": [1, 2, 3.2, 4.2]})

        self.assertListAlmostEqual(data.vary(var1=2), [1.2, 2, 3, 4])
        self.assertListAlmostEqual(data.vary(var2=2), [1, 2.2, 3, 4])
        self.assertListAlmostEqual(data.vary(var3=-2), [1, 1.8, 2.8, 4])
        self.assertListAlmostEqual(data.vary(var4=2), [1, 2, 2.6, 3.6])
            
    def test_vary_combine(self):
        """Check that vary combines multiple variations"""
        data = Yield([1, 2, 3, 4],
                     var_up={"var1": [1.1, 2, 3, 4],
                             "var2": [1, 2.1, 3, 4],
                             "var3": [1, 2.1, 3.1, 4]},
                     var_down={"var1": [0.8, 2, 3, 4],
                               "var2": [1, 1.8, 3, 4],
                               "var4": [1, 2, 3.2, 4.2]})

        self.assertListAlmostEqual(data.vary(var1=-1, var2=0.5, var3=1, var5=10),
                                   [0.8, 2.15, 3.1, 4])




import unittest
import numpy as np
from uhepp import Stack

class StackTestCase(unittest.TestCase):
    """Test the implementation of Stack"""

    def test_init_store(self):
        """Check that arguments to init are stored as members"""
        stack= Stack(["si-ggh", "si-vbf"], 'step', 'syst')

        self.assertEqual(stack.content, ['si-ggh', 'si-vbf'])
        self.assertEqual(stack.bartype, 'step')
        self.assertEqual(stack.error, 'syst')

    def test_init_default(self):
        """Check the default values for bartype and error"""
        stack= Stack(["si-ggh", "si-vbf"])

        self.assertEqual(stack.content, ['si-ggh', 'si-vbf'])
        self.assertEqual(stack.bartype, 'stepfilled')
        self.assertEqual(stack.error, 'stat')

    def test_init_copy(self):
        """Check that arguments to init are copied"""
        stackitems = ['si-ggh', 'si-vbf']
        stack = Stack(stackitems, 'step', 'syst')

        stackitems.append('si-other')

        self.assertEqual(stack.content, ['si-ggh', 'si-vbf'])
        self.assertEqual(stack.bartype, 'step')
        self.assertEqual(stack.error, 'syst')


    def test_member_change(self):
        """Check that member variables can be changed"""
        stack = Stack(["si-ggh", "si-vbf"], 'step', 'syst')

        stack.content.append('si-tth')
        stack.bartype = 'points'
        stack.error = 'env'

        self.assertEqual(stack.content, ['si-ggh', 'si-vbf', 'si-tth'])
        self.assertEqual(stack.bartype, 'points')
        self.assertEqual(stack.error, "env")

    def test_illegal_bartype(self):
        """Check that an illegal bartype raises an exception"""
        self.assertRaises(ValueError, Stack, ["si-ggh", "si-vbf"],
                          'bubble', 'syst')

        stack = Stack(["si-ggh", "si-vbf"], 'step', 'syst')
        def assign():
            stack.bartype = 'bubble'
        self.assertRaises(ValueError, assign)

    def test_illegal_error(self):
        """Check that an illegal error raises an exception"""
        self.assertRaises(ValueError, Stack, ["si-ggh", "si-vbf"],
                          'step', 'var')

        stack = Stack(["si-ggh", "si-vbf"], 'step', 'syst')
        def assign2():
            stack.error = 'var'
        self.assertRaises(ValueError, assign2)





import unittest
import numpy as np
from uhepp import RatioItem

class RatioItemTestCase(unittest.TestCase):
    """Test the implementation of RatioItem"""

    def test_init_store(self):
        """Check that arguments to init are stored as members"""
        stack = RatioItem(["ggh"], ["vbf"], 'step', 'syst',
                          linestyle=':', linewidth=3, color='r',
                          edgecolor='b')

        self.assertEqual(stack.numerator, ['ggh'])
        self.assertEqual(stack.denominator, ['vbf'])
        self.assertEqual(stack.bartype, 'step')
        self.assertEqual(stack.error, 'syst')
        self.assertEqual(stack.color, '#ff0000')

    def test_init_default(self):
        """Check the default values for bartype and error"""
        stack = RatioItem(["ggh"])

        self.assertEqual(stack.numerator, ['ggh'])
        self.assertEqual(stack.denominator, [])
        self.assertEqual(stack.bartype, 'step')
        self.assertEqual(stack.error, 'stat')
        self.assertEqual(stack.den_error, 'stat')

    def test_init_copy(self):
        """Check that arguments to init are copied"""
        stackitems = ['ggh']
        stack = RatioItem(stackitems, stackitems)

        stackitems.append('other')
        self.assertEqual(stack.numerator, ['ggh'])
        self.assertEqual(stack.denominator, ['ggh'])

    def test_member_change(self):
        """Check that member variables can be changed"""
        stack = RatioItem(["ggh"], ["vbf"])

        stack.numerator.append('tth')
        stack.denominator.append('other')
        stack.bartype = 'points'
        stack.error = 'env'
        stack.den_error = 'syst'
        stack.linewidth = 3

        self.assertEqual(stack.numerator, ['ggh', 'tth'])
        self.assertEqual(stack.denominator, ['vbf', 'other'])
        self.assertEqual(stack.bartype, 'points')
        self.assertEqual(stack.error, "env")
        self.assertEqual(stack.den_error, "syst")
        self.assertEqual(stack.linewidth, 3)




import unittest
import numpy as np
from uhepp import Line, VLine, HLine

class LineTestCase(unittest.TestCase):
    """Test the implementation of Line"""

    def test_init_store(self):
        """Check that arguments to init are stored as members"""
        line = Line(125, [0, 200], linestyle=':', linewidth=3, color='r',
                    edgecolor='b')

        self.assertEqual(line.pos, 125)
        self.assertEqual(line.stretch, [0, 200])
        self.assertEqual(line.linewidth, 3)
        self.assertEqual(line.color, '#ff0000')
        self.assertEqual(line.edgecolor, '#0000ff')

    def test_init_default(self):
        """Check the default values for bartype and error"""
        line = Line(125)

        self.assertEqual(line.pos, 125)
        self.assertIsNone(line.stretch)
        self.assertEqual(line.style, {})

    def test_member_change(self):
        """Check that member variables can be changed"""
        line = Line(125)
        line.pos = 200
        line.stretch = (0, 100)
        line.color = 'blue'

        self.assertEqual(line.pos, 200)
        self.assertEqual(line.stretch, (0, 100))
        self.assertEqual(line.color, '#0000ff')

class VLineTestCase(unittest.TestCase):
    """Test the implementation of VLine"""

    def test_init(self):
        """Check that arguments to init are stored as members"""
        line = VLine(125)
        self.assertEqual(line.pos_x, 125)

class HLineTestCase(unittest.TestCase):
    """Test the implementation of HLine"""

    def test_init(self):
        """Check that arguments to init are stored as members"""
        line = HLine(125)
        self.assertEqual(line.pos_y, 125)




import unittest
import uhepp

class OutsideTestCase(unittest.TestCase):
    """Check the code merging under- and overflow"""

    def test_output_values(self):
        """Check that under- and overflow bins are merged properly"""
        hist = uhepp.UHeppHist("x", [0, 10, 20, 30])
        hist.include_overflow = True
        hist.include_underflow = True
        hist.yields["main"] = uhepp.Yield([1, 2, 3, 4, 5])

        si = uhepp.StackItem(["main"], label="main")
        hist.stacks.append(uhepp.Stack([si], bartype='points'))
        fig, ax = hist.render()

        points = ax.lines[0]
        self.assertEqual(list(points.get_xdata()), [5, 15, 25])
        self.assertEqual(list(points.get_ydata()), [3, 3, 9])




import uhepp

def add_envelope(histo, variation, stack_index=0):
    """Returns a copy of the histogram with envelope stacks for the variation"""
    plot = histo.clone()

    stack = plot.stacks[stack_index]
    processes = [
        process
        for item in stack.content
        for process in item.yield_names
    ]

    # Up
    plot.stacks.append(uhepp.Stack([
            uhepp.StackItem(
                [f"{process}/{variation}/up" for process in processes],
                label="Up", color="#ff0000"
            )
        ], bartype="step")
    )
    plot.ratio.append(uhepp.RatioItem(
        [f"{process}/{variation}/up" for process in processes],
        processes, color="#ff0000"
    ))

    # Down
    plot.stacks.append(uhepp.Stack([
            uhepp.StackItem(
                [f"{process}/{variation}/down" for process in processes],
                label="Down", color="#0000ff"
            )
        ], bartype="step")
    )
    plot.ratio.append(uhepp.RatioItem(
        [f"{process}/{variation}/down" for process in processes],
        processes, color="#0000ff"
    ))

    return plot

def merge(first, *plots):
    """Combine multiple plots into a single plot"""
    result = first.clone()
    
    result.yields = {}
    result.stacks = []
    result.ratio = []
    for i, plot in enumerate([first, *plots]):
        plot = plot.clone()
        if plot.bin_edges != result.bin_edges:
            raise ValueError("Cannot merge incompatible bin edges")
                
        for yield_name, yield_obj in plot.yields.items():
            result.yields[f"h{i}_{yield_name}"] = yield_obj
        for stack in plot.stacks:
            for item in stack.content:
                item.yield_names = [f"h{i}_{name}"
                                    for name in item.yield_names]
            result.stacks.append(stack)
        for ratio in plot.ratio:
            ratio.numerator = [f"h{i}_{name}"
                               for name in ratio.numerator]
            ratio.denominator = [f"h{i}_{name}"
                                 for name in ratio.denominator]
            result.ratio.append(ratio)
    return result




import os
import unittest
import numpy as np
import coffea.hist as hist
from uhepp import from_yaml, from_yamls, from_json, from_jsons, from_coffea
from uhepp import from_file, UHeppHist
import uhepp

class FromTestCase(unittest.TestCase):
    """Base class to test from_XXX methods"""
    @staticmethod
    def rel_path(filename):
        """Convert a filename to a file-relative path"""
        return os.path.join(os.path.dirname(__file__), filename)

    def assertTrivialHist(self, hist, **kwds):
        """Assert the argument is the trivial toy histogram"""
        self.assertEqual(hist.version, "0.1")
        self.assertEqual(hist.date, "2020-11-10T23:39:00+01:00")
        self.assertEqual(hist.filename, "hello")
        self.assertEqual(hist.symbol, "x")
        self.assertEqual(hist.bin_edges, list(range(6)))
        self.assertEqual(hist.stacks, [])
        self.assertEqual(hist.ratio, [])
        self.assertEqual(hist.yields, {})

        self.assertIsNone(hist.author)
        self.assertIsNone(hist.lumi)
        self.assertIsNone(hist.ratio_max)

    def assertBasicHist(self, hist, **kwds):
        """Assert the argument is the basic toy histogram"""
        self.assertEqual(hist.version, "0.1")
        self.assertEqual(hist.date, "2020-11-10T23:39:00+01:00")
        self.assertEqual(hist.filename, "hello")
        self.assertEqual(hist.author, "Frank Sauerburger")
        self.assertEqual(hist.symbol, "x")
        self.assertEqual(hist.bin_edges, list(range(6)))

        stack, = hist.stacks
        self.assertEqual(stack.bartype, "points")
        content, = stack.content
        self.assertEqual(content.yield_names, ["data"])
        self.assertEqual(content.label, "Data")

        data_yield = hist.yields["data"]
        self.assertEqual(data_yield.base, [1.1, 1.9, 2.8, 4.2, 5, 6, 7])

        self.assertEqual(len(hist.yields), 1)

class FromYamlTestCase(FromTestCase):
    """Check that importing from yaml works"""

    def test_trivial_hist_string(self):
        """Check that the trivial hist is restores from a string"""
        with open(self.rel_path("trivial_hist.yaml")) as yaml_file:
            yaml_string = yaml_file.read()

        hist = from_yamls(yaml_string)
        self.assertTrivialHist(hist)

    def test_trivial_hist(self):
        """Check that importing a trivial hist restores all values"""
        hist = from_yaml(self.rel_path("trivial_hist.yaml"))
        self.assertTrivialHist(hist)

    def test_trivial_hist_render(self):
        """Check that rendering a trivial hist succeeds"""
        hist = from_yaml(self.rel_path("trivial_hist.yaml"))
        self.assertIsNotNone(hist.render())

    def test_basic_hist(self):
        """Check that importing a basic hist restores all values"""
        hist = from_yaml(self.rel_path("basic_hist.yaml"))

    def test_basic_hist_render(self):
        """Check that rendering a basic hist succeeds"""
        hist = from_yaml(self.rel_path("basic_hist.yaml"))
        self.assertIsNotNone(hist.render())

class FromFileTestCase(FromTestCase):
    """Test the auto-switching function from_file()"""

    def test_yaml_extension(self):
        """Check that a yaml file is accepted"""
        hist = from_file(self.rel_path("basic_hist.yaml"))
        self.assertIsInstance(hist, UHeppHist)

    def test_json_extension(self):
        """Check that a yaml file is accepted"""
        hist = from_file(self.rel_path("basic_hist.json"))
        self.assertIsInstance(hist, UHeppHist)

    def test_invalid_extension(self):
        """Check that a python file raises an exception"""
        self.assertRaises(ValueError, from_file, __file__)

class FromJsonTestCase(FromTestCase):
    """Check that importing from json works"""

    def test_trivial_hist_string(self):
        """Check that the trivial hist is restores from a string"""
        with open(self.rel_path("trivial_hist.json")) as json_file:
            json_string = json_file.read()

        hist = from_jsons(json_string)
        self.assertTrivialHist(hist)

    def test_trivial_hist(self):
        """Check that importing a trivial hist restores all values"""
        hist = from_json(self.rel_path("trivial_hist.json"))
        self.assertTrivialHist(hist)

    def test_trivial_hist_render(self):
        """Check that rendering a trivial hist succeeds"""
        hist = from_json(self.rel_path("trivial_hist.json"))
        self.assertIsNotNone(hist.render())

    def test_basic_hist(self):
        """Check that importing a basic hist restores all values"""
        hist = from_json(self.rel_path("basic_hist.json"))

    def test_basic_hist_render(self):
        """Check that rendering a basic hist succeeds"""
        hist = from_json(self.rel_path("basic_hist.json"))
        self.assertIsNotNone(hist.render())

class CloneTestCase(unittest.TestCase):
    """Test the implementation of clone() using to_json and from_json"""

    def test_indepenent(self):
        """Check that cloning produces an independent object"""
        hist = uhepp.UHeppHist("m", [0, 100, 200, 300])
        hist.ratio = [uhepp.RatioItem([], [])]

        new_hist = hist.clone()
        hist.ratio = []
        self.assertEqual(len(new_hist.ratio), 1)

class FromCoffeaTestCase(unittest.TestCase):
    """Test the implementation of from_coffea()"""

    @staticmethod
    def toy_coffea_hist():
        """Create and return a toy histogram  in coffea"""

        seed = int.from_bytes(b"uhepp", "big") % 2**31
        np.random.seed(seed)

        # Copy-paste example from 
        # https://coffeateam.github.io/coffea/notebooks/histograms.html

        histo = hist.Hist("Counts",
                          hist.Cat("sample", "sample name"),
                          hist.Bin("x", "x value", 20, -10, 10),
                          hist.Bin("y", "y value", 20, -10, 10),
                          hist.Bin("z", "z value", 20, -10, 10),
        )
        
        xyz = np.random.multivariate_normal(mean=[1, 3, 7], cov=np.eye(3), size=10000)
        xyz_sample2 = np.random.multivariate_normal(mean=[1, 3, 7], cov=np.eye(3), size=10000)
        weight = np.arctan(np.sqrt(np.power(xyz_sample2, 2).sum(axis=1)))

        histo.fill(sample="sample 1", x=xyz[:,0], y=xyz[:,1], z=xyz[:,2])
        histo.fill(sample="sample 2", x=xyz_sample2[:,0], y=xyz_sample2[:,1], z=xyz_sample2[:,2], weight=weight)
        
        return histo

    def test_samples(self):
        """Check that the samples are listed in the dataset return values"""
        coffea_hist = self.toy_coffea_hist()
        hists, datasets = from_coffea(coffea_hist)

        self.assertEqual(datasets, ["sample 1", "sample 2"])


    def test_yields(self):
        """Check that the yields are filled correctly"""
        coffea_hist = self.toy_coffea_hist()
        hists, datasets = from_coffea(coffea_hist)

        self.assertEqual(len(hists), 3)
        hx, hy, hz = hists

        # x projection
        self.assertEqual(hx.yields.keys(), {"sample 1", "sample 2"})
        self.assertEqual(len(hx.yields["sample 1"]), 22)
        self.assertEqual(len(hx.yields["sample 2"]), 22)

        self.assertAlmostEqual(hx.yields["sample 1"].base[9], 213)
        self.assertAlmostEqual(hx.yields["sample 1"].stat[9], np.sqrt(213))

        self.assertAlmostEqual(hx.yields["sample 2"].base[9], 324.06700678071786)
        self.assertAlmostEqual(hx.yields["sample 2"].stat[9], 21.606056455946145)

        # y projection
        self.assertEqual(hy.yields.keys(), {"sample 1", "sample 2"})
        self.assertEqual(len(hy.yields["sample 1"]), 22)
        self.assertEqual(len(hy.yields["sample 2"]), 22)

        self.assertAlmostEqual(hy.yields["sample 1"].base[12], 1377)
        self.assertAlmostEqual(hy.yields["sample 1"].stat[12], np.sqrt(1377))

        self.assertAlmostEqual(hy.yields["sample 2"].base[12], 1983.3727966421063)
        self.assertAlmostEqual(hy.yields["sample 2"].stat[12], 53.298922844614914)

        # z projection
        self.assertEqual(hz.yields.keys(), {"sample 1", "sample 2"})
        self.assertEqual(len(hz.yields["sample 1"]), 22)
        self.assertEqual(len(hz.yields["sample 2"]), 22)

        self.assertAlmostEqual(hz.yields["sample 1"].base[15], 261)
        self.assertAlmostEqual(hz.yields["sample 1"].stat[15], np.sqrt(261))

        self.assertAlmostEqual(hz.yields["sample 2"].base[15], 310.407073027837)
        self.assertAlmostEqual(hz.yields["sample 2"].stat[15], 20.834936855091176)


    def test_init_props(self):
        """Check that bins and variable are set correctly"""
        coffea_hist = self.toy_coffea_hist()
        hists, datasets = from_coffea(coffea_hist)

        self.assertEqual(len(hists), 3)
        hx, hy, hz = hists

        self.assertEqual(len(hx.bin_edges), 21)
        self.assertEqual(hx.bin_edges[0], -10)
        self.assertEqual(hx.bin_edges[-1], 10)
        self.assertEqual(hx.symbol, "x")

        self.assertEqual(len(hy.bin_edges), 21)
        self.assertEqual(hy.bin_edges[0], -10)
        self.assertEqual(hy.bin_edges[-1], 10)
        self.assertEqual(hy.symbol, "y")

        self.assertEqual(len(hz.bin_edges), 21)
        self.assertEqual(hz.bin_edges[0], -10)
        self.assertEqual(hz.bin_edges[-1], 10)
        self.assertEqual(hz.symbol, "z")

    def test_stacks(self):
        """Check that the optional stacks argument populates the stacks"""
        coffea_hist = self.toy_coffea_hist()
        hists, datasets = from_coffea(coffea_hist, stacks=True)

        self.assertEqual(len(hists), 3)
        hx, hy, hz = hists

        self.assertEqual(len(hx.stacks), 1)
        self.assertEqual(len(hx.stacks[0].content), 2)

        self.assertEqual(len(hy.stacks), 1)
        self.assertEqual(len(hy.stacks[0].content), 2)

        self.assertEqual(len(hz.stacks), 1)
        self.assertEqual(len(hz.stacks[0].content), 2)




import math
import unittest
import uhepp

class UncertaintyTestCase(unittest.TestCase):
    """Test the uncertainty computation"""

    def assertListAlmostEqual(self, a, b, *args, **kwds):
        for A, B in zip(a, b):
            self.assertAlmostEqual(A, B, *args, **kwds)
        self.assertEqual(len(a), len(b), *args, **kwds)


    def test_stat_default(self):
        """Check the uncertainty of a standard case"""
        hist = uhepp.UHeppHist("m", [100, 110, 120, 130, 140])
        hist.yields = {
            "a": uhepp.Yield([  1,   2,   3,   4,   5,   6],
                             [0.1, 0.2, 0.3, 0.4, 0.5, 0.6]),
            "b": uhepp.Yield([  2,   3,   4,   3,   2,   1],
                             [0.4, 0.4, 0.5, 0.4, 0.5, 0.5]),
            "c": uhepp.Yield([3.1, 3.2, 3.3, 3.4, 3.5, 3.6],
                             [0.1, 0.1, 0.1, 0.2, 0.2, 0.3]),
            "x": uhepp.Yield([  6,   7,   8,   9,  10,  10],
                             [0.1, 0.1, 0.1, 0.2, 0.2, 0.3])
        }

        hist.stacks.append(uhepp.Stack([
                uhepp.StackItem(["a"], label="A"),
                uhepp.StackItem(["b"], label="B"),
                uhepp.StackItem(["c"], label="C"),
        ]))
        hist.stacks.append(uhepp.Stack([
                uhepp.StackItem(["x"], label="Data"),
        ], bartype="points"))

        fig, ax = hist.render()
    
        # A polygon
        self.assertEqual(list(ax.patches[0].xy[ 0]), [100, 0])
        self.assertEqual(list(ax.patches[0].xy[ 1]), [100, 2])
        self.assertEqual(list(ax.patches[0].xy[ 2]), [110, 2])
        self.assertEqual(list(ax.patches[0].xy[ 3]), [110, 3])
        self.assertEqual(list(ax.patches[0].xy[ 4]), [120, 3])
        self.assertEqual(list(ax.patches[0].xy[ 5]), [120, 4])
        self.assertEqual(list(ax.patches[0].xy[ 6]), [130, 4])
        self.assertEqual(list(ax.patches[0].xy[ 7]), [130, 5])
        self.assertEqual(list(ax.patches[0].xy[ 8]), [140, 5])
        self.assertEqual(list(ax.patches[0].xy[ 9]), [140, 0])
        self.assertEqual(list(ax.patches[0].xy[10]), [130, 0])
        self.assertEqual(list(ax.patches[0].xy[11]), [130, 0])
        self.assertEqual(list(ax.patches[0].xy[12]), [120, 0])
        self.assertEqual(list(ax.patches[0].xy[13]), [120, 0])
        self.assertEqual(list(ax.patches[0].xy[14]), [110, 0])
        self.assertEqual(list(ax.patches[0].xy[15]), [110, 0])
        self.assertEqual(list(ax.patches[0].xy[16]), [100, 0])
        self.assertEqual(len(ax.patches[0].xy), 17)
    
        # B polygon
        self.assertEqual(list(ax.patches[1].xy[ 0]), [100, 2])
        self.assertEqual(list(ax.patches[1].xy[ 1]), [100, 5])
        self.assertEqual(list(ax.patches[1].xy[ 2]), [110, 5])
        self.assertEqual(list(ax.patches[1].xy[ 3]), [110, 7])
        self.assertEqual(list(ax.patches[1].xy[ 4]), [120, 7])
        self.assertEqual(list(ax.patches[1].xy[ 5]), [120, 7])
        self.assertEqual(list(ax.patches[1].xy[ 6]), [130, 7])
        self.assertEqual(list(ax.patches[1].xy[ 7]), [130, 7])
        self.assertEqual(list(ax.patches[1].xy[ 8]), [140, 7])
        self.assertEqual(list(ax.patches[1].xy[ 9]), [140, 5])
        self.assertEqual(list(ax.patches[1].xy[10]), [130, 5])
        self.assertEqual(list(ax.patches[1].xy[11]), [130, 4])
        self.assertEqual(list(ax.patches[1].xy[12]), [120, 4])
        self.assertEqual(list(ax.patches[1].xy[13]), [120, 3])
        self.assertEqual(list(ax.patches[1].xy[14]), [110, 3])
        self.assertEqual(list(ax.patches[1].xy[15]), [110, 2])
        self.assertEqual(list(ax.patches[1].xy[16]), [100, 2])
        self.assertEqual(len(ax.patches[1].xy), 17)
    
        # C polygon
        self.assertListAlmostEqual(list(ax.patches[2].xy[ 0]), [100,    5])
        self.assertListAlmostEqual(list(ax.patches[2].xy[ 1]), [100,  8.2])
        self.assertListAlmostEqual(list(ax.patches[2].xy[ 2]), [110,  8.2])
        self.assertListAlmostEqual(list(ax.patches[2].xy[ 3]), [110, 10.3])
        self.assertListAlmostEqual(list(ax.patches[2].xy[ 4]), [120, 10.3])
        self.assertListAlmostEqual(list(ax.patches[2].xy[ 5]), [120, 10.4])
        self.assertListAlmostEqual(list(ax.patches[2].xy[ 6]), [130, 10.4])
        self.assertListAlmostEqual(list(ax.patches[2].xy[ 7]), [130, 10.5])
        self.assertListAlmostEqual(list(ax.patches[2].xy[ 8]), [140, 10.5])
        self.assertListAlmostEqual(list(ax.patches[2].xy[ 9]), [140,    7])
        self.assertListAlmostEqual(list(ax.patches[2].xy[10]), [130,    7])
        self.assertListAlmostEqual(list(ax.patches[2].xy[11]), [130,    7])
        self.assertListAlmostEqual(list(ax.patches[2].xy[12]), [120,    7])
        self.assertListAlmostEqual(list(ax.patches[2].xy[13]), [120,    7])
        self.assertListAlmostEqual(list(ax.patches[2].xy[14]), [110,    7])
        self.assertListAlmostEqual(list(ax.patches[2].xy[15]), [110,    5])
        self.assertListAlmostEqual(list(ax.patches[2].xy[16]), [100,    5])
        self.assertEqual(len(ax.patches[2].xy), 17)
    
        # Error band
        self.assertEqual(ax.patches[3]._x0, 100)
        self.assertEqual(ax.patches[3]._width, 10)

        self.assertAlmostEqual(ax.patches[3]._height**2 / 4,
                               0.2**2 + 0.4**2 + 0.1**2)
        self.assertAlmostEqual(2 * ax.patches[3]._y0 + ax.patches[3]._height, 8.2 * 2)

        self.assertEqual(ax.patches[4]._x0, 110)
        self.assertEqual(ax.patches[3]._width, 10)
        self.assertAlmostEqual(ax.patches[4]._height**2 / 4,
                               0.3**2 + 0.5**2 + 0.1**2)
        self.assertAlmostEqual(2 * ax.patches[4]._y0 + ax.patches[4]._height, 10.3 * 2)

        self.assertEqual(ax.patches[5]._x0, 120)
        self.assertEqual(ax.patches[3]._width, 10)
        self.assertAlmostEqual(ax.patches[5]._height**2 / 4,
                               0.4**2 + 0.4**2 + 0.2**2)
        self.assertAlmostEqual(2 * ax.patches[5]._y0 + ax.patches[5]._height, 10.4 * 2)

        self.assertEqual(ax.patches[6]._x0, 130)
        self.assertEqual(ax.patches[3]._width, 10)
        self.assertAlmostEqual(ax.patches[6]._height**2 / 4,
                               0.5**2 + 0.5**2 + 0.2**2)
        self.assertAlmostEqual(2 * ax.patches[6]._y0 + ax.patches[6]._height, 10.5 * 2)

        # X points
        self.assertEqual(list(ax.lines[0].get_xdata()), [105, 115, 125, 135])
        self.assertEqual(list(ax.lines[0].get_ydata()), [7, 8, 9, 10])


        # X error
        self.assertEqual(list(ax.lines[0].get_xdata()), [105, 115, 125, 135])
        self.assertEqual(list(ax.lines[0].get_ydata()), [7, 8, 9, 10])
        self.assertEqual(len(ax.lines), 1)

        vertical = ax.containers[1][2][1].get_segments()
        self.assertEqual(list(vertical[0][0]), [105, 6.9])
        self.assertEqual(list(vertical[0][1]), [105, 7.1])
        self.assertEqual(list(vertical[1][0]), [115, 7.9])
        self.assertEqual(list(vertical[1][1]), [115, 8.1])
        self.assertEqual(list(vertical[2][0]), [125, 8.8])
        self.assertEqual(list(vertical[2][1]), [125, 9.2])
        self.assertEqual(list(vertical[3][0]), [135, 9.8])
        self.assertEqual(list(vertical[3][1]), [135, 10.2])
        self.assertEqual(len(vertical), 4)

        horizontal = ax.containers[1][2][0].get_segments()
        self.assertEqual(list(horizontal[0][0]), [100, 7])
        self.assertEqual(list(horizontal[0][1]), [110, 7])
        self.assertEqual(list(horizontal[1][0]), [110, 8])
        self.assertEqual(list(horizontal[1][1]), [120, 8])
        self.assertEqual(list(horizontal[2][0]), [120, 9])
        self.assertEqual(list(horizontal[2][1]), [130, 9])
        self.assertEqual(list(horizontal[3][0]), [130, 10])
        self.assertEqual(list(horizontal[3][1]), [140, 10])
        self.assertEqual(len(horizontal), 4)


    def test_stat_single_si(self):
        """With a single stack item using multiple yields"""
        hist = uhepp.UHeppHist("m", [100, 110, 120, 130, 140])
        hist.yields = {
            "a": uhepp.Yield([  1,   2,   3,   4,   5,   6],
                             [0.1, 0.2, 0.3, 0.4, 0.5, 0.6]),
            "b": uhepp.Yield([  2,   3,   4,   3,   2,   1],
                             [0.4, 0.4, 0.5, 0.4, 0.5, 0.5]),
            "c": uhepp.Yield([3.1, 3.2, 3.3, 3.4, 3.5, 3.6],
                             [0.1, 0.1, 0.1, 0.2, 0.2, 0.3]),
            "x": uhepp.Yield([  6,   7,   8,   9,  10,  10],
                             [0.1, 0.1, 0.1, 0.2, 0.2, 0.3])
        }

        hist.stacks.append(uhepp.Stack([
                uhepp.StackItem(["a", "b", "c"], label="A, B, C"),
        ]))
        hist.stacks.append(uhepp.Stack([
                uhepp.StackItem(["x"], label="Data"),
        ], bartype="points"))

        fig, ax = hist.render()
    
        # A polygon
        self.assertListAlmostEqual(list(ax.patches[0].xy[ 0]), [100,    0])
        self.assertListAlmostEqual(list(ax.patches[0].xy[ 1]), [100,  8.2])
        self.assertListAlmostEqual(list(ax.patches[0].xy[ 2]), [110,  8.2])
        self.assertListAlmostEqual(list(ax.patches[0].xy[ 3]), [110, 10.3])
        self.assertListAlmostEqual(list(ax.patches[0].xy[ 4]), [120, 10.3])
        self.assertListAlmostEqual(list(ax.patches[0].xy[ 5]), [120, 10.4])
        self.assertListAlmostEqual(list(ax.patches[0].xy[ 6]), [130, 10.4])
        self.assertListAlmostEqual(list(ax.patches[0].xy[ 7]), [130, 10.5])
        self.assertListAlmostEqual(list(ax.patches[0].xy[ 8]), [140, 10.5])
        self.assertListAlmostEqual(list(ax.patches[0].xy[ 9]), [140, 0])
        self.assertListAlmostEqual(list(ax.patches[0].xy[10]), [130, 0])
        self.assertListAlmostEqual(list(ax.patches[0].xy[11]), [130, 0])
        self.assertListAlmostEqual(list(ax.patches[0].xy[12]), [120, 0])
        self.assertListAlmostEqual(list(ax.patches[0].xy[13]), [120, 0])
        self.assertListAlmostEqual(list(ax.patches[0].xy[14]), [110, 0])
        self.assertListAlmostEqual(list(ax.patches[0].xy[15]), [110, 0])
        self.assertListAlmostEqual(list(ax.patches[0].xy[16]), [100, 0])
        self.assertEqual(len(ax.patches[0].xy), 17)

        # Error band
        self.assertEqual(ax.patches[1]._x0, 100)
        self.assertEqual(ax.patches[1]._width, 10)
        self.assertAlmostEqual(ax.patches[1]._height**2 / 4,
                               0.2**2 + 0.4**2 + 0.1**2)
        self.assertAlmostEqual(2 * ax.patches[1]._y0 + ax.patches[1]._height, 8.2 * 2)

        self.assertEqual(ax.patches[2]._x0, 110)
        self.assertEqual(ax.patches[2]._width, 10)
        self.assertAlmostEqual(ax.patches[2]._height**2 / 4,
                               0.3**2 + 0.5**2 + 0.1**2)
        self.assertAlmostEqual(2 * ax.patches[2]._y0 + ax.patches[2]._height, 10.3 * 2)

        self.assertEqual(ax.patches[3]._x0, 120)
        self.assertEqual(ax.patches[3]._width, 10)
        self.assertAlmostEqual(ax.patches[3]._height**2 / 4,
                               0.4**2 + 0.4**2 + 0.2**2)
        self.assertAlmostEqual(2 * ax.patches[3]._y0 + ax.patches[3]._height, 10.4 * 2)

        self.assertEqual(ax.patches[4]._x0, 130)
        self.assertEqual(ax.patches[4]._width, 10)
        self.assertAlmostEqual(ax.patches[4]._height**2 / 4,
                               0.5**2 + 0.5**2 + 0.2**2)
        self.assertAlmostEqual(2 * ax.patches[4]._y0 + ax.patches[4]._height, 10.5 * 2)


    def test_stat_rebin(self):
        """Check that uncertainty in a r3binned case"""
        hist = uhepp.UHeppHist("m", [100, 110, 120, 130, 140])
        hist.yields = {
            "a": uhepp.Yield([  1,   2,   3,   4,   5,   6],
                             [0.1, 0.2, 0.3, 0.4, 0.5, 0.6]),
            "b": uhepp.Yield([  2,   3,   4,   3,   2,   1],
                             [0.4, 0.4, 0.5, 0.4, 0.5, 0.5]),
            "c": uhepp.Yield([3.1, 3.2, 3.3, 3.4, 3.5, 3.6],
                             [0.1, 0.1, 0.1, 0.2, 0.2, 0.3]),
            "x": uhepp.Yield([  6,   7,   8,   9,  10,  10],
                             [0.1, 0.1, 0.1, 0.2, 0.2, 0.3])
        }

        hist.stacks.append(uhepp.Stack([
                uhepp.StackItem(["a"], label="A"),
                uhepp.StackItem(["b"], label="B"),
                uhepp.StackItem(["c"], label="C"),
        ]))
        hist.stacks.append(uhepp.Stack([
                uhepp.StackItem(["x"], label="Data"),
        ], bartype="points"))
        hist.rebin_edges = [100, 120, 140]

        fig, ax = hist.render()
    
        # A polygon
        self.assertEqual(list(ax.patches[0].xy[0]), [100, 0])
        self.assertEqual(list(ax.patches[0].xy[1]), [100, 5])
        self.assertEqual(list(ax.patches[0].xy[2]), [120, 5])
        self.assertEqual(list(ax.patches[0].xy[3]), [120, 9])
        self.assertEqual(list(ax.patches[0].xy[4]), [140, 9])
        self.assertEqual(list(ax.patches[0].xy[5]), [140, 0])
        self.assertEqual(list(ax.patches[0].xy[6]), [120, 0])
        self.assertEqual(list(ax.patches[0].xy[7]), [120, 0])
        self.assertEqual(list(ax.patches[0].xy[8]), [100, 0])
        self.assertEqual(len(ax.patches[0].xy), 9)
    
        # B polygon
        self.assertEqual(list(ax.patches[1].xy[0]), [100, 5])
        self.assertEqual(list(ax.patches[1].xy[1]), [100, 12])
        self.assertEqual(list(ax.patches[1].xy[2]), [120, 12])
        self.assertEqual(list(ax.patches[1].xy[3]), [120, 14])
        self.assertEqual(list(ax.patches[1].xy[4]), [140, 14])
        self.assertEqual(list(ax.patches[1].xy[5]), [140, 9])
        self.assertEqual(list(ax.patches[1].xy[6]), [120, 9])
        self.assertEqual(list(ax.patches[1].xy[7]), [120, 5])
        self.assertEqual(list(ax.patches[1].xy[8]), [100, 5])
        self.assertEqual(len(ax.patches[1].xy), 9)
    
        # C polygon
        self.assertEqual(list(ax.patches[2].xy[0]), [100, 12])
        self.assertEqual(list(ax.patches[2].xy[1]), [100, 18.5])
        self.assertEqual(list(ax.patches[2].xy[2]), [120, 18.5])
        self.assertEqual(list(ax.patches[2].xy[3]), [120, 20.9])
        self.assertEqual(list(ax.patches[2].xy[4]), [140, 20.9])
        self.assertEqual(list(ax.patches[2].xy[5]), [140, 14])
        self.assertEqual(list(ax.patches[2].xy[6]), [120, 14])
        self.assertEqual(list(ax.patches[2].xy[7]), [120, 12])
        self.assertEqual(list(ax.patches[2].xy[8]), [100, 12])
        self.assertEqual(len(ax.patches[2].xy), 9)
    
        # Error band
        self.assertEqual(ax.patches[3]._x0, 100)
        self.assertEqual(ax.patches[3]._width, 20)

        self.assertAlmostEqual(ax.patches[3]._height**2 / 4,
            0.2**2 + 0.3**2 + 0.4**2 + 0.5**2 + 0.1**2 + 0.1**2)
        self.assertAlmostEqual(2 * ax.patches[3]._y0 + ax.patches[3]._height,
                               18.5 * 2)

        self.assertEqual(ax.patches[4]._x0, 120)
        self.assertEqual(ax.patches[4]._width, 20)
        self.assertAlmostEqual(ax.patches[4]._height**2 / 4,
            0.4**2 + 0.5**2 + 0.5**2 + 0.4**2 + 0.2**2 + 0.2**2)
        self.assertAlmostEqual(2 * ax.patches[4]._y0 + ax.patches[4]._height,
                               20.9 * 2)

        # X points
        self.assertEqual(list(ax.lines[0].get_xdata()), [110, 130])
        self.assertEqual(list(ax.lines[0].get_ydata()), [15, 19])
        self.assertEqual(len(ax.lines), 1)

        # X error
        sqrt2 = math.sqrt(2)
        sqrt8 = math.sqrt(8)
        vertical = ax.containers[1][2][1].get_segments()
        self.assertListAlmostEqual(list(vertical[0][0]), [110, 15 - 0.1 * sqrt2])
        self.assertListAlmostEqual(list(vertical[0][1]), [110, 15 + 0.1 * sqrt2])
        self.assertListAlmostEqual(list(vertical[1][0]), [130, 19 - 0.1 * sqrt8])
        self.assertListAlmostEqual(list(vertical[1][1]), [130, 19 + 0.1 * sqrt8])
        self.assertEqual(len(vertical), 2)

        horizontal = ax.containers[1][2][0].get_segments()
        self.assertEqual(list(horizontal[0][0]), [100, 15])
        self.assertEqual(list(horizontal[0][1]), [120, 15])
        self.assertEqual(list(horizontal[1][0]), [120, 19])
        self.assertEqual(list(horizontal[1][1]), [140, 19])
        self.assertEqual(len(horizontal), 2)

    def test_stat_outside(self):
        """Check that uncertainty when outside bins are merged"""
        hist = uhepp.UHeppHist("m", [100, 110, 120, 130, 140])
        hist.yields = {
            "a": uhepp.Yield([  1,   2,   3,   4,   5,   6],
                             [0.1, 0.2, 0.3, 0.4, 0.5, 0.6]),
            "b": uhepp.Yield([  2,   3,   4,   3,   2,   1],
                             [0.4, 0.4, 0.5, 0.4, 0.5, 0.5]),
            "c": uhepp.Yield([3.1, 3.2, 3.3, 3.4, 3.5, 3.6],
                             [0.1, 0.1, 0.1, 0.2, 0.2, 0.3]),
            "x": uhepp.Yield([  6,   7,   8,   9,  10,  10],
                             [0.1, 0.1, 0.1, 0.2, 0.2, 0.3])
        }

        hist.stacks.append(uhepp.Stack([
                uhepp.StackItem(["a"], label="A"),
                uhepp.StackItem(["b"], label="B"),
                uhepp.StackItem(["c"], label="C"),
        ]))
        hist.stacks.append(uhepp.Stack([
                uhepp.StackItem(["x"], label="Data"),
        ], bartype="points"))

        hist.include_overflow = True
        hist.include_underflow = True

        fig, ax = hist.render()
    
        # A polygon
        self.assertEqual(list(ax.patches[0].xy[ 0]), [100, 0])
        self.assertEqual(list(ax.patches[0].xy[ 1]), [100, 3])
        self.assertEqual(list(ax.patches[0].xy[ 2]), [110, 3])
        self.assertEqual(list(ax.patches[0].xy[ 3]), [110, 3])
        self.assertEqual(list(ax.patches[0].xy[ 4]), [120, 3])
        self.assertEqual(list(ax.patches[0].xy[ 5]), [120, 4])
        self.assertEqual(list(ax.patches[0].xy[ 6]), [130, 4])
        self.assertEqual(list(ax.patches[0].xy[ 7]), [130, 11])
        self.assertEqual(list(ax.patches[0].xy[ 8]), [140, 11])
        self.assertEqual(list(ax.patches[0].xy[ 9]), [140, 0])
        self.assertEqual(list(ax.patches[0].xy[10]), [130, 0])
        self.assertEqual(list(ax.patches[0].xy[11]), [130, 0])
        self.assertEqual(list(ax.patches[0].xy[12]), [120, 0])
        self.assertEqual(list(ax.patches[0].xy[13]), [120, 0])
        self.assertEqual(list(ax.patches[0].xy[14]), [110, 0])
        self.assertEqual(list(ax.patches[0].xy[15]), [110, 0])
        self.assertEqual(list(ax.patches[0].xy[16]), [100, 0])
        self.assertEqual(len(ax.patches[0].xy), 17)
    
        # B polygon
        self.assertEqual(list(ax.patches[1].xy[ 0]), [100, 3])
        self.assertEqual(list(ax.patches[1].xy[ 1]), [100, 8])
        self.assertEqual(list(ax.patches[1].xy[ 2]), [110, 8])
        self.assertEqual(list(ax.patches[1].xy[ 3]), [110, 7])
        self.assertEqual(list(ax.patches[1].xy[ 4]), [120, 7])
        self.assertEqual(list(ax.patches[1].xy[ 5]), [120, 7])
        self.assertEqual(list(ax.patches[1].xy[ 6]), [130, 7])
        self.assertEqual(list(ax.patches[1].xy[ 7]), [130, 14])
        self.assertEqual(list(ax.patches[1].xy[ 8]), [140, 14])
        self.assertEqual(list(ax.patches[1].xy[ 9]), [140, 11])
        self.assertEqual(list(ax.patches[1].xy[10]), [130, 11])
        self.assertEqual(list(ax.patches[1].xy[11]), [130, 4])
        self.assertEqual(list(ax.patches[1].xy[12]), [120, 4])
        self.assertEqual(list(ax.patches[1].xy[13]), [120, 3])
        self.assertEqual(list(ax.patches[1].xy[14]), [110, 3])
        self.assertEqual(list(ax.patches[1].xy[15]), [110, 3])
        self.assertEqual(list(ax.patches[1].xy[16]), [100, 3])
        self.assertEqual(len(ax.patches[1].xy), 17)
    
        # C polygon
        self.assertListAlmostEqual(list(ax.patches[2].xy[ 0]), [100,    8])
        self.assertListAlmostEqual(list(ax.patches[2].xy[ 1]), [100, 14.3])
        self.assertListAlmostEqual(list(ax.patches[2].xy[ 2]), [110, 14.3])
        self.assertListAlmostEqual(list(ax.patches[2].xy[ 3]), [110, 10.3])
        self.assertListAlmostEqual(list(ax.patches[2].xy[ 4]), [120, 10.3])
        self.assertListAlmostEqual(list(ax.patches[2].xy[ 5]), [120, 10.4])
        self.assertListAlmostEqual(list(ax.patches[2].xy[ 6]), [130, 10.4])
        self.assertListAlmostEqual(list(ax.patches[2].xy[ 7]), [130, 21.1])
        self.assertListAlmostEqual(list(ax.patches[2].xy[ 8]), [140, 21.1])
        self.assertListAlmostEqual(list(ax.patches[2].xy[ 9]), [140,    14])
        self.assertListAlmostEqual(list(ax.patches[2].xy[10]), [130,    14])
        self.assertListAlmostEqual(list(ax.patches[2].xy[11]), [130,    7])
        self.assertListAlmostEqual(list(ax.patches[2].xy[12]), [120,    7])
        self.assertListAlmostEqual(list(ax.patches[2].xy[13]), [120,    7])
        self.assertListAlmostEqual(list(ax.patches[2].xy[14]), [110,    7])
        self.assertListAlmostEqual(list(ax.patches[2].xy[15]), [110,    8])
        self.assertListAlmostEqual(list(ax.patches[2].xy[16]), [100,    8])
        self.assertEqual(len(ax.patches[2].xy), 17)
    
        # Error band
        self.assertEqual(ax.patches[3]._x0, 100)
        self.assertEqual(ax.patches[3]._width, 10)

        self.assertAlmostEqual(ax.patches[3]._height**2 / 4,
            0.2**2 + 0.4**2 + 0.1**2 + 0.1**2 + 0.4**2 + 0.1**2)
        self.assertAlmostEqual(2 * ax.patches[3]._y0 + ax.patches[3]._height, 14.3 * 2)

        self.assertEqual(ax.patches[4]._x0, 110)
        self.assertEqual(ax.patches[4]._width, 10)
        self.assertAlmostEqual(ax.patches[4]._height**2 / 4,
                               0.3**2 + 0.5**2 + 0.1**2)
        self.assertAlmostEqual(2 * ax.patches[4]._y0 + ax.patches[4]._height, 10.3 * 2)

        self.assertEqual(ax.patches[5]._x0, 120)
        self.assertEqual(ax.patches[5]._width, 10)
        self.assertAlmostEqual(ax.patches[5]._height**2 / 4,
                               0.4**2 + 0.4**2 + 0.2**2)
        self.assertAlmostEqual(2 * ax.patches[5]._y0 + ax.patches[5]._height, 10.4 * 2)

        self.assertEqual(ax.patches[6]._x0, 130)
        self.assertEqual(ax.patches[6]._width, 10)
        self.assertAlmostEqual(ax.patches[6]._height**2 / 4,
            0.5**2 + 0.5**2 + 0.2**2 + 0.6**2 + 0.5**2 + 0.3**2)
        self.assertAlmostEqual(2 * ax.patches[6]._y0 + ax.patches[6]._height, 21.1 * 2)

        # X points
        self.assertEqual(list(ax.lines[0].get_xdata()), [105, 115, 125, 135])
        self.assertEqual(list(ax.lines[0].get_ydata()), [13, 8, 9, 20])

        # X error
        sqrt2 = math.sqrt(2)
        sqrt13 = math.sqrt(13)
        vertical = ax.containers[1][2][1].get_segments()
        self.assertEqual(list(vertical[0][0]), [105, 13 - 0.1 * sqrt2])
        self.assertEqual(list(vertical[0][1]), [105, 13 + 0.1 * sqrt2])
        self.assertEqual(list(vertical[1][0]), [115, 7.9])
        self.assertEqual(list(vertical[1][1]), [115, 8.1])
        self.assertEqual(list(vertical[2][0]), [125, 8.8])
        self.assertEqual(list(vertical[2][1]), [125, 9.2])
        self.assertEqual(list(vertical[3][0]), [135, 20 - 0.1 * sqrt13])
        self.assertEqual(list(vertical[3][1]), [135, 20 + 0.1 * sqrt13])
        self.assertEqual(len(vertical), 4)

        horizontal = ax.containers[1][2][0].get_segments()
        self.assertEqual(list(horizontal[0][0]), [100, 13])
        self.assertEqual(list(horizontal[0][1]), [110, 13])
        self.assertEqual(list(horizontal[1][0]), [110, 8])
        self.assertEqual(list(horizontal[1][1]), [120, 8])
        self.assertEqual(list(horizontal[2][0]), [120, 9])
        self.assertEqual(list(horizontal[2][1]), [130, 9])
        self.assertEqual(list(horizontal[3][0]), [130, 20])
        self.assertEqual(list(horizontal[3][1]), [140, 20])
        self.assertEqual(len(horizontal), 4)

    def test_stat_outside_density(self):
        """Check that uncertainty when outside bins are merged"""
        hist = uhepp.UHeppHist("m", [100, 110, 120, 130, 140])
        hist.yields = {
            "a": uhepp.Yield([  1,   2,   3,   4,   5,   6],
                             [0.1, 0.2, 0.3, 0.4, 0.5, 0.6]),
            "b": uhepp.Yield([  2,   3,   4,   3,   2,   1],
                             [0.4, 0.4, 0.5, 0.4, 0.5, 0.5]),
            "c": uhepp.Yield([3.1, 3.2, 3.3, 3.4, 3.5, 3.6],
                             [0.1, 0.1, 0.1, 0.2, 0.2, 0.3]),
            "x": uhepp.Yield([  6,   7,   8,   9,  10,  10],
                             [0.1, 0.1, 0.1, 0.2, 0.2, 0.3])
        }

        hist.stacks.append(uhepp.Stack([
                uhepp.StackItem(["a"], label="A"),
                uhepp.StackItem(["b"], label="B"),
                uhepp.StackItem(["c"], label="C"),
        ]))
        hist.stacks.append(uhepp.Stack([
                uhepp.StackItem(["x"], label="Data"),
        ], bartype="points"))

        hist.include_overflow = True
        hist.include_underflow = True
        hist.density_width = 10

        fig, ax = hist.render()
    
        # A polygon
        self.assertEqual(list(ax.patches[0].xy[ 0]), [100, 0])
        self.assertEqual(list(ax.patches[0].xy[ 1]), [100, 3])
        self.assertEqual(list(ax.patches[0].xy[ 2]), [110, 3])
        self.assertEqual(list(ax.patches[0].xy[ 3]), [110, 3])
        self.assertEqual(list(ax.patches[0].xy[ 4]), [120, 3])
        self.assertEqual(list(ax.patches[0].xy[ 5]), [120, 4])
        self.assertEqual(list(ax.patches[0].xy[ 6]), [130, 4])
        self.assertEqual(list(ax.patches[0].xy[ 7]), [130, 11])
        self.assertEqual(list(ax.patches[0].xy[ 8]), [140, 11])
        self.assertEqual(list(ax.patches[0].xy[ 9]), [140, 0])
        self.assertEqual(list(ax.patches[0].xy[10]), [130, 0])
        self.assertEqual(list(ax.patches[0].xy[11]), [130, 0])
        self.assertEqual(list(ax.patches[0].xy[12]), [120, 0])
        self.assertEqual(list(ax.patches[0].xy[13]), [120, 0])
        self.assertEqual(list(ax.patches[0].xy[14]), [110, 0])
        self.assertEqual(list(ax.patches[0].xy[15]), [110, 0])
        self.assertEqual(list(ax.patches[0].xy[16]), [100, 0])
        self.assertEqual(len(ax.patches[0].xy), 17)
    
        # B polygon
        self.assertEqual(list(ax.patches[1].xy[ 0]), [100, 3])
        self.assertEqual(list(ax.patches[1].xy[ 1]), [100, 8])
        self.assertEqual(list(ax.patches[1].xy[ 2]), [110, 8])
        self.assertEqual(list(ax.patches[1].xy[ 3]), [110, 7])
        self.assertEqual(list(ax.patches[1].xy[ 4]), [120, 7])
        self.assertEqual(list(ax.patches[1].xy[ 5]), [120, 7])
        self.assertEqual(list(ax.patches[1].xy[ 6]), [130, 7])
        self.assertEqual(list(ax.patches[1].xy[ 7]), [130, 14])
        self.assertEqual(list(ax.patches[1].xy[ 8]), [140, 14])
        self.assertEqual(list(ax.patches[1].xy[ 9]), [140, 11])
        self.assertEqual(list(ax.patches[1].xy[10]), [130, 11])
        self.assertEqual(list(ax.patches[1].xy[11]), [130, 4])
        self.assertEqual(list(ax.patches[1].xy[12]), [120, 4])
        self.assertEqual(list(ax.patches[1].xy[13]), [120, 3])
        self.assertEqual(list(ax.patches[1].xy[14]), [110, 3])
        self.assertEqual(list(ax.patches[1].xy[15]), [110, 3])
        self.assertEqual(list(ax.patches[1].xy[16]), [100, 3])
        self.assertEqual(len(ax.patches[1].xy), 17)
    
        # C polygon
        self.assertListAlmostEqual(list(ax.patches[2].xy[ 0]), [100,    8])
        self.assertListAlmostEqual(list(ax.patches[2].xy[ 1]), [100, 14.3])
        self.assertListAlmostEqual(list(ax.patches[2].xy[ 2]), [110, 14.3])
        self.assertListAlmostEqual(list(ax.patches[2].xy[ 3]), [110, 10.3])
        self.assertListAlmostEqual(list(ax.patches[2].xy[ 4]), [120, 10.3])
        self.assertListAlmostEqual(list(ax.patches[2].xy[ 5]), [120, 10.4])
        self.assertListAlmostEqual(list(ax.patches[2].xy[ 6]), [130, 10.4])
        self.assertListAlmostEqual(list(ax.patches[2].xy[ 7]), [130, 21.1])
        self.assertListAlmostEqual(list(ax.patches[2].xy[ 8]), [140, 21.1])
        self.assertListAlmostEqual(list(ax.patches[2].xy[ 9]), [140,    14])
        self.assertListAlmostEqual(list(ax.patches[2].xy[10]), [130,    14])
        self.assertListAlmostEqual(list(ax.patches[2].xy[11]), [130,    7])
        self.assertListAlmostEqual(list(ax.patches[2].xy[12]), [120,    7])
        self.assertListAlmostEqual(list(ax.patches[2].xy[13]), [120,    7])
        self.assertListAlmostEqual(list(ax.patches[2].xy[14]), [110,    7])
        self.assertListAlmostEqual(list(ax.patches[2].xy[15]), [110,    8])
        self.assertListAlmostEqual(list(ax.patches[2].xy[16]), [100,    8])
        self.assertEqual(len(ax.patches[2].xy), 17)
    
        # Error band
        self.assertEqual(ax.patches[3]._x0, 100)
        self.assertEqual(ax.patches[3]._width, 10)

        self.assertAlmostEqual(ax.patches[3]._height**2 / 4,
            0.2**2 + 0.4**2 + 0.1**2 + 0.1**2 + 0.4**2 + 0.1**2)
        self.assertAlmostEqual(2 * ax.patches[3]._y0 + ax.patches[3]._height, 14.3 * 2)

        self.assertEqual(ax.patches[4]._x0, 110)
        self.assertEqual(ax.patches[4]._width, 10)
        self.assertAlmostEqual(ax.patches[4]._height**2 / 4,
                               0.3**2 + 0.5**2 + 0.1**2)
        self.assertAlmostEqual(2 * ax.patches[4]._y0 + ax.patches[4]._height, 10.3 * 2)

        self.assertEqual(ax.patches[5]._x0, 120)
        self.assertEqual(ax.patches[5]._width, 10)
        self.assertAlmostEqual(ax.patches[5]._height**2 / 4,
                               0.4**2 + 0.4**2 + 0.2**2)
        self.assertAlmostEqual(2 * ax.patches[5]._y0 + ax.patches[5]._height, 10.4 * 2)

        self.assertEqual(ax.patches[6]._x0, 130)
        self.assertEqual(ax.patches[6]._width, 10)
        self.assertAlmostEqual(ax.patches[6]._height**2 / 4,
            0.5**2 + 0.5**2 + 0.2**2 + 0.6**2 + 0.5**2 + 0.3**2)
        self.assertAlmostEqual(2 * ax.patches[6]._y0 + ax.patches[6]._height, 21.1 * 2)

        # X points
        self.assertEqual(list(ax.lines[0].get_xdata()), [105, 115, 125, 135])
        self.assertEqual(list(ax.lines[0].get_ydata()), [13, 8, 9, 20])

        # X error
        sqrt2 = math.sqrt(2)
        sqrt13 = math.sqrt(13)
        vertical = ax.containers[1][2][1].get_segments()
        self.assertEqual(list(vertical[0][0]), [105, 13 - 0.1 * sqrt2])
        self.assertEqual(list(vertical[0][1]), [105, 13 + 0.1 * sqrt2])
        self.assertEqual(list(vertical[1][0]), [115, 7.9])
        self.assertEqual(list(vertical[1][1]), [115, 8.1])
        self.assertEqual(list(vertical[2][0]), [125, 8.8])
        self.assertEqual(list(vertical[2][1]), [125, 9.2])
        self.assertEqual(list(vertical[3][0]), [135, 20 - 0.1 * sqrt13])
        self.assertEqual(list(vertical[3][1]), [135, 20 + 0.1 * sqrt13])
        self.assertEqual(len(vertical), 4)

        horizontal = ax.containers[1][2][0].get_segments()
        self.assertEqual(list(horizontal[0][0]), [100, 13])
        self.assertEqual(list(horizontal[0][1]), [110, 13])
        self.assertEqual(list(horizontal[1][0]), [110, 8])
        self.assertEqual(list(horizontal[1][1]), [120, 8])
        self.assertEqual(list(horizontal[2][0]), [120, 9])
        self.assertEqual(list(horizontal[2][1]), [130, 9])
        self.assertEqual(list(horizontal[3][0]), [130, 20])
        self.assertEqual(list(horizontal[3][1]), [140, 20])
        self.assertEqual(len(horizontal), 4)





import unittest
import numpy as np
from uhepp import _SmartDict

class SmartDictTestCase(unittest.TestCase):
    """Test the implementation of _SmartDict"""

    def test_init(self):
        """Check that the init stores the dict"""
        sdict = _SmartDict({"hello": 42})
        self.assertEqual(sdict.data["hello"], 42)

    def test_in(self):
        """Check the in operator"""
        data = {
            "hello": {
                "world": 42,
                "this": "that",
            },
            "other": "x"
        }
        sdict = _SmartDict(data)

        self.assertIn("hello", sdict)
        self.assertIn("hello.world", sdict)
        self.assertIn("hello.this", sdict)
        self.assertIn("other", sdict)

        self.assertNotIn("world", sdict)
        self.assertNotIn("hello.other", sdict)
        self.assertNotIn("this.that", sdict)

    def test_get(self):
        """Check that get returns the default if not found"""
        data = {
            "hello": {
                "world": 42,
                "this": "that",
            },
            "other": "x"
        }
        sdict = _SmartDict(data)

        self.assertEqual(sdict.get("hello", "def"),
                        {"world": 42, "this": "that"})
        self.assertEqual(sdict.get("hello.world", "def"), 42)
        self.assertEqual(sdict.get("hello.this", "def"), "that")
        self.assertEqual(sdict.get("other", "def"), "x")

        self.assertEqual(sdict.get("world", "def"), "def")
        self.assertEqual(sdict.get("hello.other", "def"), "def")
        self.assertEqual(sdict.get("this.that", "def"), "def")

    def test_getitem(self):
        """Check that get returns the default if not found"""
        data = {
            "hello": {
                "world": 42,
                "this": "that",
            },
            "other": "x"
        }
        sdict = _SmartDict(data)

        self.assertEqual(sdict["hello"],
                         {"world": 42, "this": "that"})
        self.assertEqual(sdict["hello.world"], 42)
        self.assertEqual(sdict["hello.this"], "that")
        self.assertEqual(sdict["other"], "x")

        self.assertRaises(KeyError, lambda: sdict["world"])
        self.assertRaises(KeyError, lambda: sdict["hello.other"])
        self.assertRaises(KeyError, lambda: sdict["this.that"])



"""
This script is used to install uhepplot and all its dependencies. Run

    python setup.py install
or
    python3 setup.py install

to install the package.
"""

# Copyright (C) 2020-2021 Frank Sauerburger

from setuptools import setup

def load_long_description(filename):
    """
    Loads the given file and returns its content.
    """
    with open(filename, encoding="utf-8") as readme_file:
        content = readme_file.read()
        return content

setup(name='uhepp',
      version='0.5.0',  # Also change in module and docs
      packages=["uhepp", "uhepp.tests"],
      install_requires=["tzlocal",
                        "requests",
                        "numpy",
                        "matplotlib",
                        "atlasify",
                        "pyyaml",
                        "python-dateutil"],  # Also add in requirements.txt
      test_suite='uhepp.tests',
      scripts=['bin/uhepp'],
      description='Universal HEP plots',
      long_description=load_long_description("README.md"),
      long_description_content_type='text/markdown',
      url="https://gitlab.cern.ch/fsauerbu/uhepp",
      author="Frank Sauerburger",
      author_email="f.sauerburger@cern.ch",
      classifiers=[
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3 :: Only",
        "Topic :: Scientific/Engineering :: Physics",
      ],
      license="MIT")




# Copyright (C) 2020 Frank Sauerburger

import unittest
from uhepp import UHeppHist, Stack, StackItem, Yield

class UHepPlotingTestCase(unittest.TestCase):
    """Check the implementation of various plotting methods"""

    def test_update_style_map(self):
        """Check that all allowed values are wired to the correct output"""
        kwds= {}
        UHeppHist._update_style(kwds, {"linewidth": 1.2,
                                      "linestyle": ':',
                                      "color": (1, 2, 3),
                                      "edgecolor": (2, 2, 3)})

        self.assertEqual(kwds, {"linewidth": 1.2,
                                "linestyle": ':',
                                "color": (1, 2, 3),
                                "edgecolor": (2, 2, 3)})


    def test_update_style_missing(self):
        """Check that missing input values keep the output unchanged"""
        kwds = {"linestyle": ':', "color": (1, 2, 3)}
        UHeppHist._update_style(kwds, {})
        self.assertEqual(kwds, {"linestyle": ':', "color": (1, 2, 3)})

    def test_update_style_aux(self):
        """Check that additional input values are ignored"""
        kwds = {}
        UHeppHist._update_style(kwds, {"linestyle": ":", "whatever": 3})
        self.assertEqual(kwds, {"linestyle": ':'})

    def test_no_uncertainty(self):
        """Check that plotting succeeds without uncertainties"""
        hist = UHeppHist("m", [100, 110, 120, 130, 140])
        hist.yields = {
            "a": Yield([  1,   2,   3,   4,   5,   6],
                       [0.1, 0.2, 0.3, 0.4, 0.5, 0.6]),
            "b": Yield([  2,   3,   4,   3,   2,   1],
                       [0.4, 0.4, 0.5, 0.4, 0.5, 0.5]),
            "c": Yield([3.1, 3.2, 3.3, 3.4, 3.5, 3.6],
                       [0.1, 0.1, 0.1, 0.2, 0.2, 0.3]),
            "x": Yield([  6,   7,   8,   9,  10,  10],
                       [0.1, 0.1, 0.1, 0.2, 0.2, 0.3])
        }

        hist.stacks.append(Stack([
                StackItem(["a"], label="A"),
                StackItem(["b"], label="B"),
                StackItem(["c"], label="C"),
        ], error='no'))

        # Survival test
        hist.render()

    def test_no_stat_uncertainty(self):
        """Check that plotting succeeds without uncertainties"""
        hist = UHeppHist("m", [100, 110, 120, 130, 140])
        hist.yields = {
            "a": Yield([  1,   2,   3,   4,   5,   6]),
            "b": Yield([  2,   3,   4,   3,   2,   1]),
            "c": Yield([3.1, 3.2, 3.3, 3.4, 3.5, 3.6]),
            "x": Yield([  6,   7,   8,   9,  10,  10]),
        }

        hist.stacks.append(Stack([
                StackItem(["a"], label="A"),
                StackItem(["b"], label="B"),
                StackItem(["c"], label="C"),
        ]))

        # Survival test
        hist.render()




import unittest
import uhepp
import uhepp.utils as uu

class AddEnvelopTestCase(unittest.TestCase):
    """Test the implementation of add_envelope"""

    def setUp(self):
        """Create toy histogram"""
        self.hist = uhepp.UHeppHist("m", [0, 100, 200])

        y_bkg = uhepp.Yield([0, 1000, 500, 100],
                            var_up={"es": [0, 900, 500, 200]},
                            var_down={"es": [0, 1100, 550, 50]})
        y_sig = uhepp.Yield([0, 10, 30, 10],
                            var_up={"es": [0, 5, 35, 10]},
                            var_down={"es": [0, 10, 35, 5]})
        y_data = uhepp.Yield([0, 950, 550, 100], [0, 31, 23, 10])

        self.hist.yields = {"bkg": y_bkg, "sig": y_sig, "data": y_data}
        self.hist.stacks = [
            uhepp.Stack([
                uhepp.StackItem(["bkg"], label="Background"),
                uhepp.StackItem(["sig"], label="Background"),
            ]),
            uhepp.Stack([
                uhepp.StackItem(["data"], label="Data"),
            ], bartype="points"),
        ]
        self.hist.ratio = [
            uhepp.RatioItem(["data"], ["bkg", "sig"], bartype="points")
        ]

    def test_copy(self):
        """Check that the returned histogram is a copy"""
        env = uu.add_envelope(self.hist, "es")
        self.assertEqual(len(self.hist.stacks), 2)

        self.hist.stacks.pop()
        self.hist.stacks.pop()

        self.assertEqual(len(env.stacks), 4)

    def test_up_main(self):
        """Check the added stack item for up variation"""
        env = uu.add_envelope(self.hist, "es")

        self.assertEqual(env.stacks[2].bartype, "step")
        self.assertEqual(len(env.stacks[2].content), 1)
        self.assertEqual(env.stacks[2].content[0].yield_names,
                         ["bkg/es/up", "sig/es/up"])
        self.assertEqual(env.stacks[2].content[0].label, "Up")
        self.assertEqual(env.stacks[2].content[0].color, "#ff0000")


    def test_down_main(self):
        """Check the added stack item for down variation"""
        env = uu.add_envelope(self.hist, "es")

        self.assertEqual(env.stacks[3].bartype, "step")
        self.assertEqual(len(env.stacks[3].content), 1)
        self.assertEqual(env.stacks[3].content[0].yield_names,
                         ["bkg/es/down", "sig/es/down"])
        self.assertEqual(env.stacks[3].content[0].label, "Down")
        self.assertEqual(env.stacks[3].content[0].color, "#0000ff")

    def test_up_ratio(self):
        """Check the added ratio item for up variation"""
        env = uu.add_envelope(self.hist, "es")

        self.assertEqual(env.ratio[1].bartype, "step")
        self.assertEqual(env.ratio[1].numerator, 
                         ["bkg/es/up", "sig/es/up"])
        self.assertEqual(env.ratio[1].denominator, ["bkg", "sig"])
        self.assertEqual(env.ratio[1].color, "#ff0000")

    def test_down_ratio(self):
        """Check the added ratio item for down variation"""
        env = uu.add_envelope(self.hist, "es")

        self.assertEqual(env.ratio[2].bartype, "step")
        self.assertEqual(env.ratio[2].numerator, 
                         ["bkg/es/down", "sig/es/down"])
        self.assertEqual(env.ratio[2].denominator, ["bkg", "sig"])
        self.assertEqual(env.ratio[2].color, "#0000ff")

    def test_main_stack_index(self):
        """Check that the stack can be selected"""
        env = uu.add_envelope(self.hist, "es", stack_index=1)

        self.assertEqual(env.stacks[2].bartype, "step")
        self.assertEqual(len(env.stacks[2].content), 1)
        self.assertEqual(env.stacks[2].content[0].yield_names,
                         ["data/es/up"])

class MergeTestCase(unittest.TestCase):
    """Test the implementation of merge"""

    def setUp(self):
        """Create toy histogram"""
        self.hist_a = uhepp.UHeppHist("m", [0, 100, 200])

        y_bkg = uhepp.Yield([0, 1000, 500, 100])
        y_sig = uhepp.Yield([0, 10, 30, 10])
        y_data = uhepp.Yield([0, 950, 550, 100], [0, 31, 23, 10])

        self.hist_a.yields = {"bkg": y_bkg, "sig": y_sig, "data": y_data}
        self.hist_a.stacks = [
            uhepp.Stack([
                uhepp.StackItem(["bkg"], label="Background"),
                uhepp.StackItem(["sig"], label="Background"),
            ]),
            uhepp.Stack([
                uhepp.StackItem(["data"], label="Data"),
            ], bartype="points"),
        ]
        self.hist_a.ratio = [
            uhepp.RatioItem(["data"], ["bkg", "sig"], bartype="points")
        ]

        self.hist_b = uhepp.UHeppHist("m", [0, 100, 200])

        y_bkg = uhepp.Yield([0, 100, 50, 10])
        y_sig = uhepp.Yield([0, 1, 3, 1])
        y_data = uhepp.Yield([0, 95, 55, 10], [0, 10, 7, 3])

        self.hist_b.yields = {"bkg": y_bkg, "sig": y_sig, "data": y_data}
        self.hist_b.stacks = [
            uhepp.Stack([
                uhepp.StackItem(["bkg"], label="Background"),
                uhepp.StackItem(["sig"], label="Background"),
            ]),
            uhepp.Stack([
                uhepp.StackItem(["data"], label="Data"),
            ], bartype="points"),
        ]
        self.hist_b.ratio = [
            uhepp.RatioItem(["data"], ["bkg", "sig"], bartype="points")
        ]

    def test_merge_rename(self):
        """Check that the new yields are renamed to avoid name clash"""
        merged = uu.merge(self.hist_a, self.hist_b)

        self.assertIn("h0_bkg", merged.yields)
        self.assertIn("h0_sig", merged.yields)
        self.assertIn("h0_data", merged.yields)
        self.assertIn("h1_bkg", merged.yields)
        self.assertIn("h1_sig", merged.yields)
        self.assertIn("h1_data", merged.yields)

    def test_merge_reference(self):
        """Check that the renamed yield names are referenced"""
        merged = uu.merge(self.hist_a, self.hist_b)

        self.assertEqual(merged.stacks[0].content[0].yield_names,
                         ["h0_bkg"])
        self.assertEqual(merged.stacks[-1].content[0].yield_names,
                         ["h1_data"])

        self.assertEqual(merged.ratio[0].numerator, ["h0_data"])
        self.assertEqual(merged.ratio[-1].denominator, ["h1_bkg", "h1_sig"])

    def test_merge_complete(self):
        """Check that all the merge is complete"""
        merged = uu.merge(self.hist_a, self.hist_b)

        self.assertEqual(len(merged.stacks), 4)
        self.assertEqual(len(merged.ratio), 2)

    def test_different_binning(self):
        """Check that merge() with different bin edges raises an exception"""
        hist_b = uhepp.UHeppHist("m", [0, 100, 200, 300])

        self.assertRaises(ValueError, uu.merge, self.hist_a, hist_b)

    def test_copy(self):
        """Check that the returned histogram is a copy"""
        merged = uu.merge(self.hist_a, self.hist_b)
        self.assertEqual(len(self.hist_a.stacks), 2)

        self.hist_a.stacks.pop()
        self.hist_b.stacks.pop()

        self.assertEqual(len(merged.stacks), 4)
